#!/usr/bin/env python3
import os
import sys
import json
import timeit

import numpy as np

from snudda.neurons.neuron_prototype import NeuronPrototype
from snudda.utils.numpy_encoder import NumpyEncoder
import scipy.sparse as sparse
from scipy.spatial import distance_matrix


class SnuddaLoad(object):
    """
    Load data from network-neuron-positions.hdf5 or network-neuron-synapses.hdf5 into python dictionary
    """

    ############################################################################

    def __init__(self, network_file, snudda_data=None, load_synapses=True, verbose=False):

        """
        Constructor

        Args:
            network_file (str) : Data file to load
            snudda_data (str, optional) : Snudda Data path, if you want to override the one specified in the hdf5 file
            load_synapses (bool, optional) : Whether to read synapses into memory, or keep them on disk (this keeps file open)
            verbose (bool, optional) : Print more info during execution

        """

        # This variable will only be set if the synapses are not kept in
        # memory so that we can access them later, otherwise the hdf5 file is
        # automatically closed
        self.hdf5_file = None
        self.verbose = verbose
        self.config = None

        self.network_file = None
        self.snudda_data = snudda_data

        if network_file:
            alt_file = os.path.join(network_file, "network-synapses.hdf5")

            if os.path.isdir(network_file):
                if not os.path.isfile(alt_file):
                    raise ValueError(f"Network path {network_file} specified, but no file {alt_file}")
                network_file = alt_file

            self.data = self.load_hdf5(network_file, load_synapses)
        else:
            self.data = None

    ############################################################################

    def close(self):

        """ Close hdf5 data file. """

        if self.hdf5_file:
            self.hdf5_file.close()
            self.hdf5_file = None

    ############################################################################

    def __del__(self):

        if self.hdf5_file is not None:
            try:
                self.hdf5_file.close()
            except:
                print("Unable to close HDF5, already closed?")

    ############################################################################

    @staticmethod
    def to_str(data_str):

        """ Helper function to convert data to string. """

        if data_str is None:
            return None

        if type(data_str) in [bytes, np.bytes_]:
            return data_str.decode()

        # Warn the user if they accidentally call to_str on an int or something else
        assert type(data_str) == str, f"to_str is used on strings or bytes (that are converted to str), " \
                                      f"received {type(data_str)} -- {data_str}"

        return data_str

    ############################################################################

    def load_hdf5(self, network_file, load_synapses=True, load_morph=False):

        """
        Load data from hdf5 file.

        Args:
            network_file (str) : Network file to load data from
            load_synapses (bool) : Load synapses into memory, or read on demand from file (keeps file open)
            load_morph

        Returns:
            data (dictionary) : Dictionary with data.

    Data format. The top level of the data hierarchy are "meta", "morphologies", "network".

        "meta" hierarchy:
            "slurm_id" (int) : Run ID of Slurm job
            "axon_stump_id_flag" (bool) : Should axon be replaced by a axon stump
            "config" : Config data
            "config_file" : Config file
            "connectivity_distributions" : Information about connections allowed, including pruning information
            "hyper_voxel_ids" : List of all hyper voxels IDs
            "hyper_voxel_size" (int, int, int): Number of hyper voxels along each dimension x, y, z
            "hyper_voxel_width" (float, float, float) : Size of hypervoxel in meters (x,y,z)
            "num_hyper_voxels" (int) : Number of hyper voxels
            "position_file" : Neuron position file
            "simulation_origo" (float, float, float) : Simulation origo (x, y, z) in SI-units (m)
            "voxel_size" (float) : Voxel size in meters

        "morphologies" heirarchy:
            List of neuron names, contains the location of the morphologies
            (either SWC file or directory with multiple SWC files)

        "network" hierarchy:
            "gap_junctions" : Gap junction data matrix (see below for format)
            "num_gap_junctions" (int) : Number of gap junctions
            "num_synapses" (int) : Number of synapses
            "neurons" : Neuron data structure (see below for format)
            "synapses" : Synapse data matrix (see below for format)

    Neuron data format:


    Synapse data format (column within parenthesis):

        0: source_cell_id, 1: dest_cell_id, 2: voxel_x, 3: voxel_y, 4: voxel_z,
        5: hyper_voxel_id, 6: channel_model_id,
        7: source_axon_soma_dist (not SI scaled 1e6, micrometers),
        8: dest_dend_soma_dist (not SI scalled 1e6, micrometers)
        9: dest_seg_id, 10: dest_seg_x (int 0 - 1000, SONATA wants float 0.0-1.0)
        11: conductance (int, not SI scaled 1e12, in pS)
        12: parameterID

    Note on parameterID:

        If there are n parameter sets for the particular synapse type, then
        the ID to use is parameterID % n, this way we can reuse connectivity
        if we add more synapse parameter sets later.

    Gap junction format (column in parentheses):

        0: source_cell_id, 1: dest_cell_id, 2: source_seg_id, 3: dest_seg_id,
        4: source_seg_x, 5: dest_seg_x, 6: voxel_x, 7: voxel_y, 8: voxel_z,
        9: hyper_voxel_id, 10: conductance (integer, in pS)

        """

        assert not load_morph, "load_hdf5: load_morph=True currently disabled, does not handle morph variations"

        self.network_file = network_file

        if self.verbose:
            print(f"Loading {network_file}")

        start_time = timeit.default_timer()
        data = dict([])

        # Save a reference to the name of the loaded network file
        data["network_file"] = self.network_file

        # Blender notebook has hdf5 library/header file mismatch, so importing this only where needed
        # This allows us to use fake_load.py in snudda.utils

        import h5py
        f = h5py.File(network_file, 'r')

        # We need to keep f open if load_synapses = False, using "with" would close file

        if "config" in f["meta"]:
            if self.verbose:
                print("Loading config data from HDF5")
            data["config"] = SnuddaLoad.to_str(f["meta/config"][()])
            self.config = json.loads(f["meta/config"][()])

        # Added so this code can also load the position file, which
        # does not have the network group yet
        if "network/synapses" in f:
            try:
                data["num_neurons"] = f["network/neurons/neuron_id"].shape[0]
                data["neuron_id"] = f["network/neurons/neuron_id"][()]
            except:
                if "neuronID" in f["network/neurons"]:
                    print(f"\nERROR: Old version of the network detected!\n"
                          f"Please regenerate your Snudda network: {self.network_file}\n\n")

                import traceback
                print(traceback.format_exc())

                print("Type \"exit()\" to exit.")

                import pdb
                pdb.set_trace()

            if "num_synapses" in f["network"]:
                if f["network/num_synapses"].shape == ():
                    data["num_synapses"] = f["network/num_synapses"][()]
                else:
                    data["num_synapses"] = f["network/num_synapses"][0]

                if data["num_synapses"] != f["network/synapses"].shape[0]:
                    print(f"Expected {data['num_synapses']} synapses, found {f['network/synapses'].shape[0]} synapse rows")
                    data["num_synapses"] = f["network/synapses"].shape[0]
            else:
                data["num_synapses"] = f["network/synapses"].shape[0]

            if "num_gap_junctions" in f["network"]:
                if f["network/num_gap_junctions"].shape == ():
                    data["num_gap_junctions"] = f["network/num_gap_junctions"][()]
                else:
                    data["num_gap_junctions"] = f["network/num_gap_junctions"][0]

                if data["num_gap_junctions"] != f["network/gap_junctions"].shape[0]:
                    print(f"Expected {data['num_gap_junctions']} gap junctions, "
                          f"found {f['network/gap_junctions'].shape[0]} gap junction rows")
                    data["num_gap_junctions"] = f["network/gap_junctions"].shape[0]
            else:
                data["num_gap_junctions"] = f["network/gap_junctions"].shape[0]

            if data["num_synapses"] > 100e6:
                print(f"Found {data['num_synapses']} synapses (too many!), not loading them into memory!")
                load_synapses = False

            if "network/hyper_voxel_ids" in f:
                data["hyper_voxel_ids"] = f["network/hyper_voxel_ids"][()]

            if load_synapses:
                # 0: source_cell_id, 1: dest_cell_id, 2: voxel_x, 3: voxel_y, 4: voxel_z,
                # 5: hyper_voxel_id, 6: channel_model_id,
                # 7: source_axon_soma_dist (not SI scaled 1e6, micrometers),
                # 8: dest_dend_soma_dist (not SI scalled 1e6, micrometers)
                # 9: dest_seg_id, 10: dest_seg_x (int 0 - 1000, SONATA wants float 0.0-1.0)
                # 11: conductance (int, not SI scaled 1e12, in pS)
                # 12: parameterID
                #
                # Note on parameterID:
                # If there are n parameter sets for the particular synapse type, then
                # the ID to use is parameterID % n, this way we can reuse connectivity
                # if we add more synapse parameter sets later.

                data["synapses"] = f["network/synapses"][:]
                data["gap_junctions"] = f["network/gap_junctions"][:]

                # !!! Convert from voxel idx to coordinates
                if f["network/synapses"].shape[0] > 0:
                    data["synapse_coords"] = f["network/synapses"][:, 2:5] * f["meta/voxel_size"][()] \
                                            + f["meta/simulation_origo"][()]
                else:
                    data["synapse_coords"] = np.zeros((3, 0))
            else:
                # Point the data structure to the synapses and gap junctions on file
                # This will be slower, and only work while the file is open
                data["synapses"] = f["network/synapses"]
                data["gap_junctions"] = f["network/gap_junctions"]

                # We need to keep f alive, since we did not load synapses into
                # the memory
                self.hdf5_file = f

        else:
            data["num_neurons"] = f["network/neurons/neuron_id"].shape[0]
            assert data["num_neurons"] == f["network/neurons/neuron_id"][-1] + 1, \
                "Internal error, something fishy with number of neurons found"

        data["config_file"] = SnuddaLoad.to_str(f["meta/config_file"][()])

        if "meta/position_file" in f:
            data["position_file"] = SnuddaLoad.to_str(f["meta/position_file"][()])

        if "meta/slurm_id" in f:
            if type(f["meta/slurm_id"][()]) in [bytes, np.bytes_]:
                data["slurm_id"] = int(f["meta/slurm_id"][()].decode())
            else:
                data["slurm_id"] = int(f["meta/slurm_id"][()])

        else:
            if self.verbose:
                print("No slurm_id set, using -1")
            data["slurm_id"] = -1

        if "meta/simulation_origo" in f:
            data["simulation_origo"] = f["meta/simulation_origo"][()]

        if "meta/voxel_size" in f:
            data["voxel_size"] = f["meta/voxel_size"][()]

        if "meta/axon_stump_id_flag" in f:
            data["axon_stump_id_flag"] = f["meta/axon_stump_id_flag"][()]

        if "meta/snudda_data" in f:
            data["snudda_data"] = SnuddaLoad.to_str(f["meta/snudda_data"][()])

            if self.snudda_data is None:
                self.snudda_data = data["snudda_data"]

        data["neurons"] = self.extract_neurons(f)

        # This is for old format, update for new format
        if "parameters" in f:
            # print("Parameters found, loading")
            data["synapse_range"] = f["parameters/synapse_range"][()]
            data["gap_junction_range"] = f["parameters/gap_junction_range"][()]
            data["min_synapse_spacing"] = f["parameters/min_synapse_spacing"][()]

        data["neuron_positions"] = f["network/neurons/position"][()]
        data["name"] = [SnuddaLoad.to_str(x) for x in f["network/neurons/name"][()]]

        if "population_unit_id" in f["network/neurons"]:
            data["population_unit"] = f["network/neurons/population_unit_id"][()]
        else:
            if self.verbose:
                print("No Population Units detected.")
            data["population_unit"] = np.zeros(data["num_neurons"], dtype=int)

        # TODO: Remove this, or make it able to handle multiple morphologies for each neuron_name,
        #  ie when morphologies is given as a dir
        if load_morph and "morphologies" in f:
            data["morph"] = dict([])

            for morph_name in f["morphologies"].keys():
                data["morph"][morph_name] = {"swc": f["morphologies"][morph_name]["swc"][()],
                                             "location": f["morphologies"][morph_name]["location"][()]}

        data["connectivity_distributions"] = dict([])

        if "connectivity_distributions" in f["meta"]:
            orig_connectivity_distributions = \
                json.loads(SnuddaLoad.to_str(f["meta/connectivity_distributions"][()]))

            for keys in orig_connectivity_distributions:
                (pre_type, post_type) = keys.split("$$")
                data["connectivity_distributions"][pre_type, post_type] \
                    = orig_connectivity_distributions[keys]

        if "synapses" in data and self.verbose:
            if "gap_junctions" in data:
                print(f"Loading {len(data['neurons'])} neurons with {data['num_synapses']} synapses"
                      f" and {data['num_gap_junctions']} gap junctions")
            else:
                print(f"Loading {len(data['neurons'])} neurons with {data['synapses'].shape[0]} synapses")

        if self.verbose:
            print(f"Load done. {timeit.default_timer() - start_time:.1f}")

        if load_synapses:
            f.close()
        else:
            self.hdf5_file = f

        return data

    ############################################################################

    @staticmethod
    def gather_extra_axons(hdf5_file):

        extra_axons = dict()

        if "extra_axons" in hdf5_file["network/neurons"]:
            for neuron_id, axon_name, position, rotation, swc_file \
                    in zip(hdf5_file["network/neurons/extra_axons/parent_neuron"],
                           hdf5_file["network/neurons/extra_axons/name"],
                           hdf5_file["network/neurons/extra_axons/position"],
                           hdf5_file["network/neurons/extra_axons/rotation"],
                           hdf5_file["network/neurons/extra_axons/morphology"]):

                if neuron_id not in extra_axons:
                    extra_axons[neuron_id] = dict()

                extra_axons[neuron_id][axon_name] = dict()
                extra_axons[neuron_id][axon_name]["position"] = position.copy()
                extra_axons[neuron_id][axon_name]["rotation"] = rotation.copy().reshape(3, 3)
                extra_axons[neuron_id][axon_name]["morphology"] = SnuddaLoad.to_str(swc_file)

        return extra_axons

    ############################################################################

    @staticmethod
    def extract_neurons(hdf5_file):

        """
        Helper function to extract neuron data from hdf5 file and put it in a dictionary.

        Args:
            hdf5_file : hdf5 file object

        Returns:
            List containing neurons as dictionary elements.

        """

        neurons = []
        extra_axons = SnuddaLoad.gather_extra_axons(hdf5_file=hdf5_file)

        for name, neuron_id, hoc, pos, rot, virtual, vID, \
            axon_density_type, axon_density, axon_density_radius, \
            axon_density_bounds_xyz, \
            morph, neuron_path, \
            parameter_key, morphology_key, modulation_key, population_unit_id \
                in zip(hdf5_file["network/neurons/name"][:],
                       hdf5_file["network/neurons/neuron_id"][:],
                       hdf5_file["network/neurons/hoc"][:],
                       hdf5_file["network/neurons/position"][()],
                       hdf5_file["network/neurons/rotation"][()],
                       hdf5_file["network/neurons/virtual_neuron"][:],
                       hdf5_file["network/neurons/volume_id"][:],
                       hdf5_file["network/neurons/axon_density_type"][:],
                       hdf5_file["network/neurons/axon_density"][:],
                       hdf5_file["network/neurons/axon_density_radius"][:],
                       hdf5_file["network/neurons/axon_density_bounds_xyz"][:],
                       hdf5_file["network/neurons/morphology"][:],
                       hdf5_file["network/neurons/neuron_path"][:],
                       hdf5_file["network/neurons/parameter_key"][:],
                       hdf5_file["network/neurons/morphology_key"][:],
                       hdf5_file["network/neurons/modulation_key"][:],
                       hdf5_file["network/neurons/population_unit_id"][:]
                       ):

            n = dict([])

            n["name"] = SnuddaLoad.to_str(name)

            if morph is not None:
                n["morphology"] = SnuddaLoad.to_str(morph)

            # Naming convention is TYPE_X, where XX is a number starting from 0
            n["type"] = n["name"].split("_")[0]

            n["neuron_id"] = neuron_id
            n["volume_id"] = SnuddaLoad.to_str(vID)
            n["hoc"] = SnuddaLoad.to_str(hoc)
            n["neuron_path"] = SnuddaLoad.to_str(neuron_path)

            n["position"] = pos.copy()
            n["rotation"] = rot.copy().reshape(3, 3)
            n["virtual_neuron"] = virtual

            if len(axon_density_type) > 0:
                n["axon_density_type"] = SnuddaLoad.to_str(axon_density_type)
            else:
                n["axon_density_type"] = None

            if len(axon_density) > 0:
                n["axon_density"] = SnuddaLoad.to_str(axon_density)
            else:
                n["axon_density"] = None

            if n["axon_density_type"] == "xyz":
                n["axon_density_bounds_xyz"] = axon_density_bounds_xyz
            else:
                n["axon_density_bounds_xyz"] = None

            n["axon_density_radius"] = axon_density_radius

            # If the code fails here, use snudda/utils/upgrade_old_network_file.py to upgrade your old data files
            par_key = SnuddaLoad.to_str(parameter_key)
            morph_key = SnuddaLoad.to_str(morphology_key)
            mod_key = SnuddaLoad.to_str(modulation_key)
            n["parameter_key"] = par_key if len(par_key) > 0 else None
            n["morphology_key"] = morph_key if len(morph_key) > 0 else None
            n["modulation_key"] = mod_key if len(mod_key) > 0 else None

            n["population_unit"] = population_unit_id

            if neuron_id in extra_axons:
                n["extra_axons"] = extra_axons[neuron_id].copy()

            neurons.append(n)

        return neurons

    ############################################################################

    def load_config_file(self):

        """ Load config data from JSON file. """

        if self.config is None:
            config_file = self.data["config_file"]
            self.config = json.load(open(config_file, 'r'))

    ############################################################################

    def synapse_iterator(self, chunk_size=1000000, data_type=None):

        """
        Iterates through all synapses in chunks (default 1e6 synapses).

        Args:
            chunk_size (int) : Number of synapses per chunk
            data_type (string) : "synapses" (default) or "gap_junctions"

        Returns:
            Iterator over the synapses
        """

        if data_type is None:
            data_type = "synapses"

        # data_type is "synapses" or "gap_junctions"
        assert data_type in ["synapses", "gap_junctions"]

        num_rows = self.data[data_type].shape[0]
        if num_rows == 0:
            # No synapses
            return

        chunk_size = min(num_rows, chunk_size)
        num_steps = int(np.ceil(num_rows / chunk_size))

        row_start = 0

        for row_end in np.linspace(chunk_size, num_rows, num_steps, dtype=int):
            synapses = self.data[data_type][row_start:row_end, :]
            row_start = row_end

            yield synapses

    ############################################################################

    def gap_junction_iterator(self, chunk_size=1000000):

        """
        Iterates through all gap junctions in chunks (default 1e6 gap junctions).

        Args:
            chunk_size (int) : Number of gap junctions per chunk

        Returns:
            Iterator over the gap junctions
        """

        return self.synapse_iterator(chunk_size=chunk_size, data_type="gap_junctions")

    ############################################################################

    # Helper methods for sorting

    @staticmethod
    def _row_eval_post_pre(row, num_neurons):
        return row[1] * num_neurons + row[0]

    @staticmethod
    def _row_eval_post(row, num_neurons):
        return row[1]

    ############################################################################

    def find_synapses_slow(self, pre_id, n_max=1000000):

        """
        Returns subset of synapses.

        Args:
            pre_id (int) : Pre-synaptic neuron ID (can also be a list)
            n_max (int) : Maximum number of synapses to return

        Returns:
            Subset of synapse matrix, synapse coordinates

        """

        if self.verbose:
            print(f"Finding synapses originating from {pre_id}, this is slow")

        synapses = np.zeros((n_max, 13), dtype=np.int32)
        syn_ctr = 0

        if np.issubdtype(type(pre_id), np.integer):
            for syn_list in self.synapse_iterator():
                for syn in syn_list:
                    if syn[0] == pre_id:
                        synapses[syn_ctr, :] = syn
                        syn_ctr += 1

        else:
            for syn_list in self.synapse_iterator():
                for syn in syn_list:
                    if syn[0] in pre_id:
                        synapses[syn_ctr, :] = syn
                        syn_ctr += 1

        synapse_coords = synapses[:, 2:5][:syn_ctr, :] * self.data["voxel_size"] + self.data["simulation_origo"]

        return synapses[:syn_ctr, :], synapse_coords

    ############################################################################

    # Either give preID and postID, or just postID

    def find_synapses(self, pre_id=None, post_id=None, silent=True, return_index=False):

        """
        Returns subset of synapses.

        Args:
            pre_id (int) : Pre-synaptic neuron ID
            post_id (int) : Post-synaptic neuron ID
            silent (bool) : Work quietly or verbosely

        Returns:
            Subset of synapse matrix.

        """

        if self.data["synapses"].shape[0] == 0:
            if not silent:
                print("No synapses in network")
            if return_index:
                return None, None, None
            else:
                return None, None

        if post_id is None:
            assert return_index is False, "You must specify pre_id and post_id if return_index is True"
            return self.find_synapses_slow(pre_id=pre_id)

        assert post_id is not None, "Must specify at least postID"

        num_rows = self.data["synapses"].shape[0]
        num_neurons = len(self.data["neuron_id"])

        if pre_id is None:
            row_eval = self._row_eval_post
            val_target = post_id
        else:
            row_eval = self._row_eval_post_pre
            val_target = post_id * num_neurons + pre_id

        idx_a1 = 0
        idx_a2 = num_rows - 1

        idx_found = None

        # We use idxA1 and idxA2 as upper and lower range within which we
        # hope to find one of the synapses. Once we found a synapse row
        # we go up and down in matrix to find the range of the synapses
        # matching the requested condition. This works because the matrix is
        # sorted on postID, and then preID if postID matches

        if row_eval(self.data["synapses"][idx_a1, :], num_neurons) == val_target:
            idx_found = idx_a1

        if row_eval(self.data["synapses"][idx_a2, :], num_neurons) == val_target:
            idx_found = idx_a2

        # -1 since if idx_a1 and idx_a2 are one apart, we have checked all values
        while idx_a1 < idx_a2 - 1 and idx_found is None:

            idx_next = int(np.round((idx_a1 + idx_a2) / 2))
            val_next = row_eval(self.data["synapses"][idx_next, :], num_neurons)

            if val_next < val_target:
                idx_a1 = idx_next
            elif val_next > val_target:
                idx_a2 = idx_next
            else:
                # We found a hit
                idx_found = idx_next
                break

        if idx_found is None:
            # No synapses found
            if self.verbose:
                print("No synapses found")

            if return_index:
                return None, None, None
            else:
                return None, None

        # Find start of synapse range
        idx_b1 = idx_found
        val_b1 = row_eval(self.data["synapses"][idx_b1 - 1, :], num_neurons)

        while val_b1 == val_target and idx_b1 > 0:
            idx_b1 -= 1
            val_b1 = row_eval(self.data["synapses"][idx_b1 - 1, :], num_neurons)

        # Find end of synapse range
        idx_b2 = idx_found

        if idx_b2 + 1 < self.data["synapses"].shape[0]:
            val_b2 = row_eval(self.data["synapses"][idx_b2 + 1, :], num_neurons)

            while val_b2 == val_target and idx_b2 + 1 < self.data["synapses"].shape[0]:
                idx_b2 += 1
                val_b2 = row_eval(self.data["synapses"][idx_b2 + 1, :], num_neurons)

        synapses = self.data["synapses"][idx_b1:idx_b2 + 1, :].copy()

        if not silent and self.verbose:
            print(f"Synapse range, first {idx_b1}, last {idx_b2}")
            print(f"{synapses}")

        # Calculate coordinates
        synapse_coords = synapses[:, 2:5] * self.data["voxel_size"] + self.data["simulation_origo"]

        if return_index:
            return synapses, synapse_coords, np.arange(idx_b1, idx_b2+1)
        else:
            return synapses, synapse_coords

    ############################################################################

    def get_neuron_population_units(self, neuron_id=None, return_set=False):

        if neuron_id is not None:
            neuron_population_units = self.data["population_unit"][neuron_id].flatten().copy()
        else:
            neuron_population_units = self.data["population_unit"].flatten().copy()

        if return_set:
            return set(neuron_population_units)
        else:
            return neuron_population_units

    def get_neuron_types(self, neuron_id=None, return_set=False):

        if neuron_id is not None:
            neuron_types = [self.data["neurons"][x]["type"] for x in neuron_id]
        else:
            neuron_types = [x["type"] for x in self.data["neurons"]]

        if return_set:
            return set(neuron_types)
        else:
            return neuron_types

    ############################################################################

    # Returns neuron_id of all neurons of neuron_type
    # OBS, random_permute is not using a controled rng, so not affected by random seed set

    def get_neuron_id_of_type(self, neuron_type, num_neurons=None, random_permute=False, volume=None,
                              include_virtual=True):

        """
        Find all neuron ID of a specific neuron type.

        Args:
            neuron_type (string) : Neuron type (e.g. "FS")
            num_neurons (int) : Maximum number of neurons to return
            random_permute (bool) : Shuffle the resulting neuron IDs?
            volume (string) : volume_id containing neurons (default None -- all neurons of type)

        Returns:
            List of neuron ID of specified neuron type

        """

        neuron_id = np.array([x["neuron_id"] for x in self.data["neurons"]
                              if (neuron_type is None or x["type"] == neuron_type)
                              and (volume is None or x["volume_id"] == volume)
                              and (include_virtual or not x["virtual_neuron"])])

        assert not random_permute or num_neurons is not None, "random_permute is only valid when num_neurons is given"

        if num_neurons is not None:
            if random_permute:
                # Do not use this if you have a simulation with multiple
                # workers... they might randomize differently, and you might
                # fewer neurons in total than you wanted
                keep_idx = np.random.permutation(len(neuron_id))

                if len(keep_idx) > num_neurons:
                    keep_idx = keep_idx[:num_neurons]

                neuron_id = neuron_id[keep_idx]
            else:
                neuron_id = neuron_id[:num_neurons]

            if len(neuron_id) < num_neurons:
                if self.verbose:
                    print(f"get_neuron_id_of_type: wanted {num_neurons} only got {len(neuron_id)} "
                          f"neurons of type {neuron_type}")

        # Double check that all of the same type (or neuron_type is None)
        assert neuron_type is None or np.array([self.data["neurons"][x]["type"] == neuron_type for x in neuron_id]).all()
        assert volume is None or np.array([self.data["neurons"][x]["volume_id"] == volume for x in neuron_id]).all()

        return neuron_id

    def get_neuron_id(self, include_virtual=True):

        """ Returns all neuron_id, if include_virtual is set (default) virtual neurons are also included."""

        neuron_id = np.array([x["neuron_id"] for x in self.data["neurons"] if include_virtual or not x["virtual_neuron"]])

        return neuron_id

    def get_neuron_id_with_name(self, neuron_name, include_virtual=True):

        """
        Find neuron ID of neurons with a given name.

        Args:
            neuron_name (str): Name of neurons (e.g. "dSPN_0")
            include_virtual (bool): Should virtual neurons also be included?

        Returns:
            List of neuron ID
        """

        neuron_id = np.array([x["neuron_id"] for x in self.data["neurons"]
                              if x["name"] == neuron_name and (include_virtual or not x["virtual_neuron"])])

        return neuron_id

    def get_population_unit_members(self, population_unit, num_neurons=None, random_permute=False):

        """
        Returns neuron ID of neurons belonging to a specific population unit.

        Args:
            population_unit (int) : Population unit ID
            num_neurons (int) : Number of neurons to return (None = all hits)
            random_permute (bool) : Randomly shuffle neuron IDs?

        Returns:
            List of neuron ID belonging to population unit.

        """

        neuron_id = np.where(self.data["population_unit"] == population_unit)[0]

        if num_neurons:
            if random_permute:
                neuron_id = np.random.permutation(neuron_id)

            if len(neuron_id) > num_neurons:
                neuron_id = neuron_id[:num_neurons]

        # Just double check
        assert (self.data["population_unit"][neuron_id] == population_unit).all()

        return neuron_id

    def load_neuron(self, neuron_id):

        """
        Loads a specific neuron. Returns a NeuronMorphology object.

        Args:
            neuron_id (int): Neuron ID

        Returns:
            NeuronMorphology object.

        """

        neuron_prototype = NeuronPrototype(neuron_path=self.data["neurons"][neuron_id]["neuron_path"],
                                           snudda_data=self.data["snudda_data"],
                                           neuron_name=self.data["neurons"][neuron_id]["name"])
        neuron_object = neuron_prototype.clone(parameter_key=self.data["neurons"][neuron_id]["parameter_key"],
                                               morphology_key=self.data["neurons"][neuron_id]["morphology_key"],
                                               modulation_key=self.data["neurons"][neuron_id]["modulation_key"],
                                               position=self.data["neurons"][neuron_id]["position"],
                                               rotation=self.data["neurons"][neuron_id]["rotation"])
        return neuron_object

    def iter_neuron_id(self):

        for x, v in enumerate(self.data["neurons"]):
            assert x == v["neuron_id"], \
                f"Neuron at position {x} has neuron_id {v['neuron_id']} (should be same)"
            yield x

    def get_neuron_keys(self, neuron_id):
        n = self.data["neurons"][neuron_id]
        return n["parameter_key"], n["morphology_key"], n["modulation_key"]

    def get_neuron_params(self, neuron_id):

        neuron_path = self.data["neurons"][neuron_id]["neuron_path"]
        parameter_key = self.data["neurons"][neuron_id]["parameter_key"]
        parameter_file = os.path.join(neuron_path, "parameters.json")
        with open(parameter_file, "r") as f:
            parameter_data = json.load(f)

        param_data = parameter_data[parameter_key]

        if "modulation_key" in self.data["neurons"][neuron_id]:
            modulation_key = self.data["neurons"][neuron_id]["modulation_key"]
            modulation_file = os.path.join(neuron_path, "modulation.json")
            with open(modulation_file, "r") as f:
                modulation_data = json.load(f)
            mod_data = modulation_data[modulation_key]
        else:
            mod_data = None

        return param_data, mod_data

    def find_gap_junctions(self, neuron_id, n_max=1000000, return_index=False):

        """ Find gap junctions associated with neuron_id

        Args:
            neuron_id (int) : Neuron with gap junction (can also be a list)
            n_max (int) : Maximum number of gap junctions to return
            return_index (bool): Should third return value index be present

        Returns:
            Subset of gap junction matrix, gap junction coordinates

        """

        if self.verbose:
            print(f"Finding gap junctions connecting neuron {neuron_id}")

        gap_junctions = np.zeros((n_max, 11), dtype=np.int32)
        gj_ctr = 0
        gj_index = 0
        gj_index_list = np.zeros((n_max,), dtype=int)

        if np.issubdtype(type(neuron_id), np.integer):
            for gj_list in self.gap_junction_iterator():
                for gj in gj_list:
                    if gj[0] == neuron_id or gj[1] == neuron_id:
                        gap_junctions[gj_ctr, :] = gj
                        gj_ctr += 1
                        gj_index_list[gj_ctr] = gj_index
                    gj_index += 1
        else:
            for gj_list in self.gap_junction_iterator():
                for gj in gj_list:
                    if gj[0] in neuron_id or gj[1] in neuron_id:
                        gap_junctions[gj_ctr, :] = gj
                        gj_ctr += 1
                        gj_index_list[gj_ctr] = gj_index
                    gj_index += 1

        gj_coords = gap_junctions[:, 6:9][:gj_ctr, :] * self.data["voxel_size"] + self.data["simulation_origo"]

        if return_index:
            return gap_junctions[:gj_ctr, :], gj_coords, gj_index_list[:gj_ctr]
        else:
            return gap_junctions[:gj_ctr, :], gj_coords

    def get_centre_neurons_iterator(self, n_neurons=None, neuron_type=None, centre_point=None, max_distance=None,
                                    return_distance=True, include_virtual=False):

        """ Return neuron id:s, starting from the centre most and moving outwards

        Args:
            n_neurons (int) : Number of neurons to return, None = all available
            neuron_type (str) : Type of neurons to return, None = all available
            centre_point (np.array) : x,y,z of centre position, None = auto detect centre
        """

        if centre_point is None:
            centre_point = np.mean(self.data["neuron_positions"], axis=0)

        dist_to_centre = np.linalg.norm(self.data["neuron_positions"] - centre_point, axis=-1)
        idx = np.argsort(dist_to_centre)

        neuron_ctr = 0

        for neuron_id in idx:
            if not include_virtual and self.data["neurons"][neuron_id]["virtual_neuron"]:
                # Ignore virtual neurons
                continue

            if neuron_type is not None and self.data["neurons"][neuron_id]["type"] != neuron_type:
                # Wrong neuron type
                continue

            if max_distance is not None and dist_to_centre[neuron_id] > max_distance:
                # Stop iterator if max distance is reached
                return

            if return_distance:
                yield neuron_id, dist_to_centre[neuron_id]
            else:
                yield neuron_id

            neuron_ctr += 1

            if n_neurons is not None and neuron_ctr >= n_neurons:
                # Stop iterator if n_neurons are delivered
                return

    ############################################################################

    def create_connection_matrix(self, sparse_matrix=True):

        if sparse_matrix:
            connection_matrix = sparse.lil_matrix((self.data["num_neurons"], self.data["num_neurons"]), dtype=np.int16)
        else:
            connection_matrix = np.zeros((self.data["num_neurons"], self.data["num_neurons"]), dtype=np.int16)

        for syn_row in self.data["synapses"]:
            connection_matrix[syn_row[0], syn_row[1]] += 1

        return connection_matrix

    def create_distance_matrix(self, neuron_id=None, pre_id=None, post_id=None):

        if neuron_id is not None and pre_id is not None and post_id is not None:
            raise ValueError("Specify either neuron_id or the two parameters pre_id and post_id.")

        if (pre_id is None) ^ (post_id is None):
            raise ValueError("pre_id and post_id must both either be specified, or neither")

        pos = self.data["neuron_positions"]

        if neuron_id is not None:
            pos = pos[neuron_id, :]
            dist_matrix = distance_matrix(pos, pos)

        elif pre_id is not None and post_id is not None:
            dist_matrix = distance_matrix(pos[pre_id, :], pos[post_id, :])

        else:
            dist_matrix = distance_matrix(pos, pos)

        return dist_matrix

    def print_all_synapse_counts_per_type(self):

        synapse_types = sorted(list(self.get_neuron_types(return_set=True)))
        connection_matrix = self.create_connection_matrix()

        for pre_type in synapse_types:
            for post_type in synapse_types:
                count = self.count_synapses_per_type(pre_neuron_type=pre_type,
                                                     post_neuron_type=post_type,
                                                     connection_matrix=connection_matrix)
                if count > 0:
                    print(f"{pre_type} -> {post_type}: {count} synapses")

    def count_synapses_per_type(self, pre_neuron_type=None, post_neuron_type=None, connection_matrix=None):

        """
            Args:
                pre_neuron_type (list): List of neuron types
                post_neuron_type (list): List of neuron types, e.g. ["dSPN", "iSPN"]
                connection_matrix

        """

        if connection_matrix is None:
            connection_matrix = self.create_connection_matrix()

        pre_mask = np.zeros((connection_matrix.shape[0]), dtype=bool)
        post_mask = np.zeros((connection_matrix.shape[1]), dtype=bool)

        if pre_neuron_type is None:
            pre_neuron_type = self.get_neuron_types(return_set=True)
        elif type(pre_neuron_type) != list:
            pre_neuron_type = [pre_neuron_type]

        if post_neuron_type is None:
            post_neuron_type = self.get_neuron_types(return_set=True)
        elif type(post_neuron_type) != list:
            post_neuron_type = [post_neuron_type]

        for neuron_type in pre_neuron_type:
            pre_id = self.get_neuron_id_of_type(neuron_type)
            pre_mask[pre_id] = True

        for neuron_type in post_neuron_type:
            post_id = self.get_neuron_id_of_type(neuron_type)
            post_mask[post_id] = True

        synapse_count = np.sum(connection_matrix[pre_id, :][:, post_id])

        return synapse_count

    def count_incoming_connections(self, neuron_type):

        neuron_id = self.get_neuron_id_of_type(neuron_type)
        neuron_id_mask = np.zeros((self.data["num_neurons"],), dtype=bool)
        neuron_id_mask[neuron_id] = True

        synapse_count = 0
        gap_junction_count = 0

        for synapses in self.synapse_iterator():
            synapse_count += np.sum(neuron_id_mask[synapses[:, 1]])

        for gap_junctions in self.gap_junction_iterator():
            gap_junction_count += np.sum(neuron_id_mask[gap_junctions[:, 0]])
            gap_junction_count += np.sum(neuron_id_mask[gap_junctions[:, 1]])

        gap_junction_count /= 2

        return synapse_count, gap_junction_count


def snudda_load_cli():
    """ Command line parser for SnuddaLoad script """

    from argparse import ArgumentParser

    parser = ArgumentParser(description="Load snudda network file (hdf5)")
    parser.add_argument("network_file", help="Network file (hdf5)", type=str)
    parser.add_argument("--listN", help="Lists neurons in network", action="store_true")
    parser.add_argument("--listT", type=str, help="List neurons of type, --listT ? list the types.", default=None)
    parser.add_argument("--listPre", help="List pre synaptic neurons", type=int)
    parser.add_argument("--listPost", help="List post synaptic neurons (slow)", type=int)
    parser.add_argument("--listGJ", help="List gap junctions (slow)", type=int)
    parser.add_argument("--listTotalIncoming", help="List number of total incoming connections to neuron type",
                        type=str, default=None)
    parser.add_argument("--keepOpen", help="This prevents loading of synapses to memory, and keeps HDF5 file open",
                        action="store_true")
    parser.add_argument("--detailed", help="More information", action="store_true")
    parser.add_argument("--voxels", help="Voxel information", action="store_true")
    parser.add_argument("--centre", help="List n neurons in centre (-1 = all)", type=int)
    parser.add_argument("--listParam", help="List parameters for neuron_id", type=int)
    parser.add_argument("--countSyn", help="Count synapses per type", action="store_true")

    args = parser.parse_args()

    if args.keepOpen:
        load_synapses = False
    else:
        load_synapses = True

    nl = SnuddaLoad(args.network_file, load_synapses=load_synapses, verbose=True)

    if args.listN:
        print("Neurons in network: ")

        if args.detailed:
            for nid, name, pos, par_key, morph_key, mod_key, neuron_path, pop_id \
                    in [(x["neuron_id"], x["name"], x["position"],
                         x["parameter_key"], x["morphology_key"], x["modulation_key"], x["neuron_path"], x["population_unit"])
                        for x in nl.data["neurons"]]:
                print("%d : %s  (x: %f, y: %f, z: %f) pop id %d, par_key: %s, morph_key: %s, mod_key: %s, neuron_path: %s"
                      % (nid, name, pos[0], pos[1], pos[2], pop_id, par_key, morph_key, mod_key, neuron_path))
        else:
            for nid, name, pos, pid in [(x["neuron_id"], x["name"], x["position"], x["population_unit"]) for x in nl.data["neurons"]]:
                print("%d : %s [%d] (x: %f, y: %f, z: %f)" % (nid, name, pid, pos[0], pos[1], pos[2]))

    if args.listT is not None:
        if args.listT == "?":
            print("List neuron types in network:")

            n_types = np.unique([x["type"] for x in nl.data["neurons"]])
            for nt in n_types:
                num = len([x["type"] for x in nl.data["neurons"] if x["type"] == nt])
                print(f"{nt} ({num} total)")

        else:
            print(f"Neurons of type {args.listT}:")
            n_of_type = [(x["neuron_id"], x["name"]) for x in nl.data["neurons"]
                         if x["type"] == args.listT]
            for nid, name in n_of_type:
                print("%d : %s" % (nid, name))

    if args.listPre is not None:
        print(f"List neurons pre-synaptic to neuron_id = {args.listPre} "
              f"({nl.data['neurons'][args.listPre]['name']})")
        synapses, synapse_coords = nl.find_synapses(post_id=args.listPre)

        if synapses is None:
            print("No pre synaptic neurons were found.")
        else:
            print(f"The neuron receives {synapses.shape[0]} synapses")
            pre_id = np.unique(synapses[:, 0])

            for nid, name in [(x["neuron_id"], x["name"]) for x in nl.data["neurons"] if x["neuron_id"] in pre_id]:
                n_syn = np.sum(synapses[:, 0] == nid)
                print(f"{nid} : {name} ({n_syn} synapses)")

                if args.detailed:
                    idx = np.where(synapses[:, 0] == nid)[0]
                    for i in idx:
                        print(f" -- SecID {synapses[i, 9]}, SecX {synapses[i, 10] * 1e-3:.2f}, "
                              f"Soma dist: {synapses[i,8]:.1f} μm, "
                              f"Coord: ({synapse_coords[i, 0]*1e6:.1f}, "
                              f"{synapse_coords[i, 1]*1e6:.1f}, "
                              f"{synapse_coords[i, 2]*1e6:.1f}) μm, "
                              f"Cond: {synapses[i, 11] * 1e-3:.2f} nS")

                    print("")

                if args.voxels:
                    idx = np.where(synapses[:, 0] == nid)[0]
                    for i in idx:
                        print(f" -- voxels: {synapses[i, 2:5]}, hyper id: {synapses[i, 5]}, "
                              f" sec id {synapses[i, 9]}, sec x {synapses[i, 10] * 1e-3:.2f}, "
                              f"Soma dist: {synapses[i,8]:.1f} μm, ")
                    print("")

    if args.listPost is not None:
        print(f"List neurons post-synaptic to neuron_id = {args.listPost}"
              f" ({nl.data['neurons'][args.listPost]['name']}):")
        synapses, synapse_coords = nl.find_synapses(pre_id=args.listPost)
        print(f"The neuron makes {synapses.shape[0]} synapses on other neurons")

        if synapses is None:
            print("No post synaptic targets found.")
        else:
            post_id = np.unique(synapses[:, 1])

            for nid, name in [(x["neuron_id"], x["name"]) for x in nl.data["neurons"] if x["neuron_id"] in post_id]:
                n_syn = np.sum(synapses[:, 1] == nid)
                print(f"{nid} : {name} ({n_syn} synapses)")

                if args.detailed:
                    idx = np.where(synapses[:, 1] == nid)[0]
                    for i in idx:
                        print(f" -- SecID {synapses[i, 9]}, SecX {synapses[i, 10] * 1e-3:.2f}, "
                              f"Soma dist: {synapses[i,8]:.1f} μm, "
                              f"Coord: ({synapse_coords[i, 0]*1e6:.1f}, "
                              f"{synapse_coords[i, 1]*1e6:.1f}, "
                              f"{synapse_coords[i, 2]*1e6:.1f}) μm, "
                              f"Cond: {synapses[i, 11] * 1e-3:.2f} nS")

                    print("")

    if args.listGJ is not None:
        print(f"List gap junctions of neuron_id = {args.listGJ}"
              f" ({nl.data['neurons'][args.listGJ]['name']})")
        gap_junctions, gap_junction_coords = nl.find_gap_junctions(neuron_id=args.listGJ)

        if gap_junctions.shape[0] == 0:
            print("No gap junctions on neuron.")
        else:
            connected_id = set(gap_junctions[:, 0]).union(gap_junctions[:, 1])
            connected_id.remove(args.listGJ)

            for nid, name in [(x["neuron_id"], x["name"]) for x in nl.data["neurons"] if x["neuron_id"] in connected_id]:
                n_gj = np.sum(gap_junctions[:, 0] == nid) + np.sum(gap_junctions[:, 1] == nid)
                print(f"{nid} : {name} ({n_gj} gap junctions)")

                if args.detailed:
                    idx1 = np.where(gap_junctions[:, 0] == nid)[0]
                    for i in idx1:
                        print(f" -- SecID {gap_junctions[i, 2]}, SecX {gap_junctions[i, 4] * 1e-3:.3f}, "
                              f"Coord: ({gap_junction_coords[i, 0]*1e6:.1f}, "
                              f"{gap_junction_coords[i, 1]*1e6:.1f}, "
                              f"{gap_junction_coords[i, 2]*1e6:.1f}) μm, "
                              f"Cond: {gap_junctions[i, 10] * 1e-3:.3f} nS")

                    idx2 = np.where(gap_junctions[:, 1] == nid)[0]
                    for i in idx2:
                        print(f" -- SecID {gap_junctions[i, 3]}, SecX {gap_junctions[i, 5] * 1e-3:.3f}, "
                              f"Coord: ({gap_junction_coords[i, 0]*1e6:.1f}, "
                              f"{gap_junction_coords[i, 1]*1e6:.1f}, "
                              f"{gap_junction_coords[i, 2]*1e6:.1f}) μm, "
                              f"Cond: {gap_junctions[i, 10] * 1e-3:.3f} nS")

    if args.listParam is not None:
        param_data, mod_data = nl.get_neuron_params(neuron_id=args.listParam)
        print(f"Neuron ID {args.listParam}\n"
              f"parameters: {json.dumps(param_data, indent=4, cls=NumpyEncoder)}\n\n"
              f"modulation: {json.dumps(mod_data, indent=4, cls=NumpyEncoder)}\n")

    if args.centre:
        if args.centre < 0:
            n_neurons = None
        else:
            n_neurons = args.centre

        for neuron_id, centre_dist in nl.get_centre_neurons_iterator(n_neurons):
            pos = nl.data["neurons"][neuron_id]["position"] * 1e6
            name = nl.data["neurons"][neuron_id]["name"]
            print(f"{neuron_id} {name}: ({pos[0]:.1f}, {pos[1]:.1f}, {pos[2]:.1f}) μm, distance to centre {centre_dist*1e6:.1f} μm")

    if args.countSyn:
        nl.print_all_synapse_counts_per_type()

    if args.listTotalIncoming:
        incoming_to_type = args.listTotalIncoming

        synapse_count, gap_junction_count = nl.count_incoming_connections(neuron_type=incoming_to_type)
        print(f"All neurons of type {incoming_to_type} receive in total {synapse_count:.0f} synapses, "
              f"and have {gap_junction_count:.0f} gap junctions in total.")


if __name__ == "__main__":
    snudda_load_cli()



#!/usr/bin/env python3
import json
import os.path
import sys
from collections import OrderedDict

import h5py
import numpy as np

from snudda.utils.load import SnuddaLoad


class SnuddaAblateNetwork:
    """ Ablate neurons or synapses from a network.

        For an example config see Snudda/examples/config/example-ablation-config.json

          {
            "ablate_neurons": ["FS",
		       ["iSPN", 0.3],
		       [1,2,3,4]],
            "ablate_synapses" : [["dSPN", "iSPN"],
			 ["iSPN", "dSPN", 0.1]]
          }

        Here "ablate_neurons" is a list of neurons to remove. The list can contain neuron types (e.g. "FS"),
        or neuron type and removal probability (e.g. ["iSPN", 0.3], or neuron ID to remove (e.g. [1,2,3,4]).

        With "ablate_synapses" the user can specify what synapses to remove (e.g. ["dSPN", "iSPN"]] to remove
        all synapses between dSPN and iSPN). Alternatively the removal probability can also be specified
        (e.g. ["iSPN", "dSPN", 0.1] where 10% of connections between those neuron types are removed).


    """

    def __init__(self, network_file, verbose=False):

        self.snudda_load = SnuddaLoad(network_file=network_file, load_synapses=False, verbose=verbose)
        self.in_file = self.snudda_load.hdf5_file
        self.h5libver = "latest"
        self.h5driver = "sec2"

        self.keep_neuron_id = None
        self.removed_connection_type = None
        self.remove_pair_connection_list = None
        self.remove_all_synapses = False
        self.remove_all_gap_junctions = False
        self.make_virtual_neuron_id = None

        self.reset_network()

    def reset_network(self):

        """ Marks all neurons to be kept. """

        self.keep_neuron_id = set(self.in_file["network/neurons/neuron_id"][:])
        self.removed_connection_type = []
        self.remove_pair_connection_list = []
        self.make_virtual_neuron_id = []

        self.remove_all_synapses = False
        self.remove_all_gap_junctions = False

    def ablate_all_synapses(self):
        self.remove_all_synapses = True

    def ablate_all_gap_junctions(self):
        self.remove_all_gap_junctions = True

    def remove_neuron_id(self, neuron_id):

        print(f"Removing neuron_id={set(neuron_id)}")
        self.keep_neuron_id = self.keep_neuron_id - set(neuron_id)

    def only_keep_neuron_id(self, neuron_id):
        self.keep_neuron_id = set(neuron_id)

    def make_virtual(self, neuron_id):

        if type(neuron_id) == list:
            self.make_virtual_neuron_id += neuron_id
        elif isinstance(neuron_id, np.ndarray):
            self.make_virtual_neuron_id += [x for x in neuron_id]
        else:
            self.make_virtual_neuron_id.append(neuron_id)

    def remove_neuron_type(self, neuron_type, p_remove=1):

        """ Remove neuron of type neuron_type with probability p_remove (default 1)"""

        remove_cell_id = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type)
        remove_flag = np.random.uniform(size=(len(remove_cell_id),)) <= p_remove
        remove_cell_id = remove_cell_id[remove_flag]

        if len(remove_flag) > 0:
            print(f"Marking {neuron_type} ({np.sum(remove_flag)} out of {len(remove_flag)}) for removal (P={p_remove})")
        else:
            available_neuron_types = sorted(list(set([x["type"] for x in self.snudda_load.data["neurons"]])))
            print(f"No {neuron_type} found in network. Available types are {', '.join(available_neuron_types)}")

        self.keep_neuron_id = self.keep_neuron_id - set(remove_cell_id)

    def remove_neuron_name(self, neuron_name):

        """ Remove neuron with name neuron_name """

        remove_cell_id = self.snudda_load.get_neuron_id_with_name(neuron_name=neuron_name)

        if len(remove_cell_id) > 0:
            print(f"Marking {neuron_name} ({len(remove_cell_id)}) for removal")
        else:
            available_neuron_names = sorted(list(set([x["name"] for x in self.snudda_load.data["neurons"]])))
            print(f"No {neuron_name} found in network. Available types are {', '.join(available_neuron_names)}")

        self.keep_neuron_id = self.keep_neuron_id - set(remove_cell_id)

    def remove_connection(self, pre_neuron_type, post_neuron_type, p_remove=1):

        """ Removes connections between specified neuron types. """

        available_neuron_types = sorted(list(set([x["type"] for x in self.snudda_load.data["neurons"]])))
        if pre_neuron_type not in available_neuron_types or post_neuron_type not in available_neuron_types:
            print(f"ERROR: Bad connection type {pre_neuron_type},{post_neuron_type}\n"
                  f"Available neuron types: {available_neuron_types}")
            return

        print(f"Marking {pre_neuron_type}, {post_neuron_type} synapses for removal (P={p_remove}).")
        self.removed_connection_type.append((pre_neuron_type, post_neuron_type, p_remove))

    def remove_pair_connection(self, pre_id, post_id):
        self.remove_pair_connection_list.append((pre_id, post_id))

    def filter_synapses(self, data_type):

        """ Filters synapses, data_type is either 'synapses' or 'gap_junctions' """

        synapse_data = self.in_file[f"network/{data_type}"][()].copy()

        if synapse_data.size == 0:
            return np.array([], dtype=int)

        keep_flag = np.zeros((synapse_data.shape[0],), dtype=bool)

        # Shortcut, if user wants to remove all, skip the processing part
        if data_type == "synapses" and self.remove_all_synapses:
            return keep_flag  # All zero
        elif data_type == "gap_junctions" and self.remove_all_gap_junctions:
            return keep_flag

        neuron_types = [n["type"] for n in self.snudda_load.data["neurons"]]

        prev_source = None
        prev_dest = None
        prev_status = None

        n_original_synapses = synapse_data.shape[0]

        for idx, (pre_id, post_id) in enumerate(zip(synapse_data[:, 0], synapse_data[:, 1])):

            if idx % 10000000 == 0:
                print(f"{idx}/{n_original_synapses} synapses processed")

            if pre_id == prev_source and post_id == prev_dest:
                keep_flag[idx] = prev_status
            else:
                prev_source = pre_id
                prev_dest = post_id

                if pre_id in self.keep_neuron_id and post_id in self.keep_neuron_id:

                    row_status = 1

                    if self.removed_connection_type:
                        for con_type in self.removed_connection_type:
                            if neuron_types[pre_id] == con_type[0] and neuron_types[post_id] == con_type[1]:
                                if np.random.uniform() <= con_type[2]:
                                    # All synapses between a given neuron pair is removed together
                                    row_status = 0
                                    break

                    keep_flag[idx] = row_status
                    prev_status = row_status
                else:
                    prev_status = 0

        print(f"{n_original_synapses}/{n_original_synapses} synapses processed")

        if len(self.remove_pair_connection_list) > 0:
            print(f"Warning, removing individual synapses ({len(self.remove_pair_connection_list)}) can be slow, "
                  f"use this with caution.")
            for pre_id, post_id in self.remove_pair_connection_list:
                if data_type == "synapses":
                    _, _, synapse_idx = self.snudda_load.find_synapses(pre_id=pre_id, post_id=post_id, return_index=True)
                elif data_type == "gap_junctions":
                    print(f"WARNING: If there are any gap junctions between {pre_id}, {post_id} they will not have been removed ")
                    # TODO: There is currently only find_gap_junctions that take neuron_id as parameter, not pre and post id
                    #       so we can not find the gap junction pairs, this will need to be added in the future.
                    # _, _, synapse_idx = self.snudda_load.find_gap_junctions(pre_id=pre_id, post_id=post_id, return_index=True)
                    synapse_idx = None
                else:
                    assert f"Unknown data type: {data_type}, should be 'synapses' or 'gap_junctions'"

                if synapse_idx is not None:
                    for syn_idx in synapse_idx:
                        keep_flag[syn_idx] = False

        print("Filtering done.")

        return keep_flag

    def write_remapping_file(self, remap_file_name):

        with open(remap_file_name, "w") as f:
            for new_id, old_id in enumerate(sorted(list(self.keep_neuron_id))):
                f.write(f"{old_id}, {new_id}\n")

    def write_network(self, out_file_name=None, print_remapping=False):

        """ Write network to hdf5 file: output_file_name """

        if not out_file_name:
            out_file_name = f"{self.in_file.filename}-modified.hdf5"

        assert out_file_name != self.in_file.filename, f"In and out file must be different."

        print(f"Writing to {out_file_name}")
        out_file = h5py.File(out_file_name, "w", libver=self.h5libver, driver=self.h5driver)

        if "config" in self.in_file:
            self.in_file.copy("config", out_file)

        self.in_file.copy("meta", out_file)

        if "morphologies" in self.in_file:
            print("Copying morphologies")
            self.in_file.copy("morphologies", out_file)

        soma_keep_id = sorted(list(self.keep_neuron_id))
        num_soma_keep = len(soma_keep_id)

        print(f"Keeping {num_soma_keep} neurons.")

        # We need to remap neuron_id in the synapses and gap junction matrix
        remap_id = np.full((len(self.snudda_load.data["neurons"]),), np.nan, dtype=int)
        for new_id, old_id in enumerate(soma_keep_id):
            remap_id[old_id] = new_id

        if print_remapping:
            print("\nRemapping neurons:")
            for new_id, old_id in enumerate(soma_keep_id):
                assert remap_id[old_id] == new_id, f"Internal error with remap_id"
                if old_id == new_id:
                    print(f"{old_id} the same")
                else:
                    print(f"{old_id} -> {new_id}")
            print("")

        network_group = out_file.create_group("network")
        neuron_group = network_group.create_group("neurons")

        for var_name in self.in_file["network/neurons"]:

            if var_name == "extra_axons":

                parent_id = self.in_file["network/neurons/extra_axons/parent_neuron"][()]

                # Identify which axons belong to neurons that are kept
                keep_axon_id = np.where([p in soma_keep_id for p in parent_id])[0]
                axon_dummy_remapping = np.arange(len(keep_axon_id))
                axon_group = neuron_group.create_group("extra_axons")

                for var_name2 in self.in_file["network/neurons/extra_axons"]:
                    self.copy_neuron_item(neuron_group=axon_group,
                                          var_name=var_name2,
                                          soma_keep_id=keep_axon_id,
                                          remap_id=axon_dummy_remapping,
                                          neuron_path="network/neurons/extra_axons")
            else:
                self.copy_neuron_item(neuron_group=neuron_group,
                                      var_name=var_name,
                                      soma_keep_id=soma_keep_id,
                                      remap_id=remap_id)

                if var_name == "virtual_neuron" and len(self.make_virtual_neuron_id) > 0:
                    virt_id = [remap_id[x] for x in self.make_virtual_neuron_id]
                    neuron_group[var_name][virt_id] = True

                    virt_neuron_names = [self.snudda_load.data["neurons"][x]["name"] for x in self.make_virtual_neuron_id]

                    for old_nrn_id, new_virt_id, v_name in zip(self.make_virtual_neuron_id, virt_id, virt_neuron_names):
                        print(f"Making neuron id {new_virt_id} ({v_name}) virtual (old ID {old_nrn_id})")

        if "synapses" in self.in_file["network"]:

            # Next deal with synapses
            keep_syn_flag = self.filter_synapses(data_type="synapses")

            # Lastly deal with gap junctions
            keep_gj_flag = self.filter_synapses(data_type="gap_junctions")

            num_syn = np.sum(keep_syn_flag)
            num_synapses = np.zeros((1,), dtype=np.uint64) + num_syn

            num_gj = np.sum(keep_gj_flag)
            num_gap_junctions = np.zeros((1,), dtype=np.uint64) + num_gj

            network_group.create_dataset("num_synapses", data=num_synapses, dtype=np.uint64)
            network_group.create_dataset("num_gap_junctions", data=num_gap_junctions,
                                         dtype=np.uint64)

            # TODO: !!!! might need to handle chunk size differently based on size...

            syn_mat = self.in_file["network/synapses"][()].copy()
            gj_mat = self.in_file["network/gap_junctions"][()].copy()

            print("Copying synapses and gap junctions")

            n_synapses = syn_mat.shape[0]

            temp_syn_mat = np.zeros((num_syn, syn_mat.shape[1]), dtype=np.int32)
            temp_gj_mat = np.zeros((num_gj, gj_mat.shape[1]), dtype=np.int32)

            syn_keep_idx = np.where(keep_syn_flag)[0]
            for idx, row_idx in enumerate(syn_keep_idx):

                if idx % 50000 == 0 and idx > 0:
                    print(f"{idx} / {num_syn} synapse rows parsed")

                # We need to remap the neuron_id if some neurons have been removed!!
                row = syn_mat[row_idx, :]
                row[0] = remap_id[row[0]]
                row[1] = remap_id[row[1]]
                temp_syn_mat[idx, :] = row

            network_group.create_dataset("synapses",
                                         data=temp_syn_mat,
                                         dtype=np.int32, shape=(num_syn, self.in_file["network/synapses"].shape[1]),
                                         chunks=self.in_file["network/synapses"].chunks,
                                         maxshape=(None, self.in_file["network/synapses"].shape[1]),
                                         compression=self.in_file["network/synapses"].compression)

            print(f"{num_syn} / {num_syn} synapse rows parsed")
            print("Synapse matrix written.")

            print(f"Keeping {num_syn} synapses (out of {syn_mat.shape[0]})")

            n_gj = gj_mat.shape[0]

            for idx, row_idx in enumerate(np.where(keep_gj_flag)[0]):

                if idx % 50000 == 0:
                    print(f"{idx} / {num_gj} gap junction rows parsed")

                # We need to remap the neuron_id if some neurons have been removed!!
                row = gj_mat[row_idx, :]
                row[0] = remap_id[row[0]]
                row[1] = remap_id[row[1]]
                temp_gj_mat[idx, :] = row

            if temp_gj_mat.size > 0:
                gj_chunk_size = self.in_file["network/gap_junctions"].chunks
            else:
                gj_chunk_size = None

            network_group.create_dataset("gap_junctions",
                                         data=temp_gj_mat,
                                         dtype=np.int32, shape=(num_gj, self.in_file["network/gap_junctions"].shape[1]),
                                         chunks=gj_chunk_size,
                                         maxshape=(None, self.in_file["network/gap_junctions"].shape[1]),
                                         compression=self.in_file["network/gap_junctions"].compression)

            print(f"{num_gj} / {num_gj} gap junction rows parsed")
            print("Gap junction matrix written.")
            print(f"Keeping {num_gj}  gap junctions (out of {gj_mat.shape[0]})")

        else:
            print("No synapses found (assuming this was a save file with only position information).")

        out_file.close()

        remapping_file = f"{out_file_name}-remapping.txt"
        self.write_remapping_file(remapping_file)

    def copy_neuron_item(self, neuron_group, var_name, soma_keep_id, remap_id, neuron_path="network/neurons"):

        data = self.in_file[f"{neuron_path}/{var_name}"]
        num_soma_keep = len(soma_keep_id)

        if len(data.shape) == 0:
            # Scalar data, just copy
            self.in_file.copy(f"{neuron_path}/{var_name}", neuron_group)
            return

        elif len(data.shape) == 1:
            # 1D data, we only keep nSomaKeep of them
            data_shape = (num_soma_keep,)
        elif len(data.shape) == 2:
            # 2D data, need to make sure to maintain dimensions
            data_shape = (num_soma_keep, data.shape[1])
        else:
            raise ValueError("copy_item: Only handle 0D, 1D and 2D data, update code!")

        if var_name == "neuron_id":
            # We need to remap
            neuron_group.create_dataset(var_name, data_shape, data.dtype,
                                        [remap_id[data[x]] for x in soma_keep_id],
                                        compression=data.compression)

            # Double check that it is OK, should be in order after
            assert (np.diff(neuron_group["neuron_id"][()]) == 1).all(), "Problem with neuron remapping!"

        else:
            try:
                neuron_group.create_dataset(var_name, data_shape, data.dtype,
                                            [data[x] for x in soma_keep_id],
                                            compression=data.compression)
            except:
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()


def snudda_ablate_network_cli():
    from argparse import ArgumentParser

    # TODO: Fix so ablation can be specified using a json file for more complex ablations

    parser = ArgumentParser(description="Modify connections in network.")
    parser.add_argument("original_network", type=str, help="Input network hdf5 file")
    parser.add_argument("output_network", type=str, help="Output network hdf5 file", default=None)
    parser.add_argument("--config", type=str, help="Ablation config file", default=None)
    parser.add_argument("--remove_neuron_type", type=str, help="Neuron type to remove", default=None)
    parser.add_argument("--p_remove_neuron_type", type=float, help="Probability to remove neuron of type", default=1.0)
    parser.add_argument("--remove_neuron_name", type=str, help="Neuron name to remove", default=None)
    parser.add_argument("--remove_neuron_id", type=str, help="Neuron ID to remove (e.g. 4,5,6)", default=None)
    parser.add_argument("--remove_connection", type=str, help="Connection to remove (e.g. 'dSPN','iSPN'", default=None)
    parser.add_argument("--p_remove_connection", type=float, help="Probability to remove connection", default=1.0)
    parser.add_argument("--make_virtual", type=str, help="Neuron ID to make virtual (e.g. 1,2,3)")
    args = parser.parse_args()

    mod_network = SnuddaAblateNetwork(network_file=args.original_network)

    if args.config:
        if not os.path.isfile(args.config):
            print(f"There is no config file {args.config}")
            exit(-1)

        with open(args.config) as f:
            config_data = json.load(f, object_pairs_hook=OrderedDict)

        if "ablate_neurons" in config_data:
            for ablate in config_data["ablate_neurons"]:
                if type(ablate[0]) == int:
                    # Ablate neurons with number
                    mod_network.remove_neuron_id(ablate)
                elif "_" in ablate[0]:
                    # Ablate neuron with name
                    mod_network.remove_neuron_name(neuron_name=ablate[0])
                else:
                    neuron_type = ablate[0]
                    if type(ablate) == list and len(ablate) > 1:
                        ablate_p = float(ablate[1])
                    else:
                        ablate_p = 1

                    mod_network.remove_neuron_type(neuron_type=neuron_type, p_remove=ablate_p)

            for con in config_data["ablate_synapses"]:
                pre_type = con[0]
                post_type = con[1]
                if len(con) > 2:
                    p_remove = con[2]
                else:
                    p_remove = 1
                mod_network.remove_connection(pre_neuron_type=pre_type, post_neuron_type=post_type, p_remove=p_remove)

    if args.remove_neuron_type:
        mod_network.remove_neuron_type(neuron_type=args.remove_neuron_type,
                                       p_remove=args.p_remove_neuron_type)

    if args.remove_neuron_name:
        mod_network.remove_neuron_name(neuron_name=args.remove_neuron_name)

    if args.remove_neuron_id:
        neuron_id = [int(x) for x in args.remove_neuron_id.split(",")]
        mod_network.remove_neuron_id(neuron_id=neuron_id)

    if args.make_virtual:
        v_id = [int(x) for x in args.make_virtual.split(",")]
        mod_network.make_virtual(neuron_id=v_id)

    if args.remove_connection:

        if args.remove_connection.count(";") > 0:

            for c in args.remove_connection.split(";"):
                assert args.remove_connection.count(",") == 1, \
                    "Format is --remove_connection pre_neuron_type, post_neuron_type"

                pre_type, post_type = c.split(",")
                mod_network.remove_connection(pre_neuron_type=pre_type, post_neuron_type=post_type,
                                              p_remove=args.p_remove_connection)


        else:
            assert args.remove_connection.count(",") == 1, \
                "Format is --remove_connection pre_neuron_type,post_neuron_type"

            pre_type, post_type = args.remove_connection.split(",")
            mod_network.remove_connection(pre_neuron_type=pre_type, post_neuron_type=post_type,
                                          p_remove=args.p_remove_connection)

    mod_network.write_network(out_file_name=args.output_network)


if __name__ == "__main__":
    snudda_ablate_network_cli()



import json
import os
import numpy as np
import copy


class Neuromodulation:

    def __init__(self):

        self.network_wide = dict()
        self.name_to_key = dict()
        self.dt = None
        self.type = 'replay'

    def set_modulation(self, neurotransmitter, neurotransmitter_key):

        """
        neurotransmitter_key is equivalent to the parameter used in the mod files, which marks the
        level, mod and maxMod parameters eg. neurotransmitter_key, ACh, would have parameters levelACh,
        modACh and maxModACh in modulated mod files.
        """

        if neurotransmitter_key in self.network_wide.keys():
            raise KeyError('neurotransmitter already defined')

        else:
            self.name_to_key.update({neurotransmitter: neurotransmitter_key})

            self.network_wide.update({neurotransmitter_key: {'name': neurotransmitter}})

            self.network_wide[neurotransmitter_key].update({'ion_channels': dict(),
                                                            'receptors': dict(),
                                                            'presynaptic': dict()})

    def transient(self, neurotransmitter, method, duration, parameters):

        self.network_wide[self.name_to_key[neurotransmitter]].update({'method': method,
                                                                      'duration': duration,
                                                                      'parameters': parameters})

    def set_timestep(self, dt):

        self.dt = dt

    def ion_channel_modulation(self, neurotransmitter, cell_type, section, ion_channels):

        if cell_type not in self.network_wide[self.name_to_key[neurotransmitter]]['ion_channels'].keys():
            self.network_wide[self.name_to_key[neurotransmitter]]['ion_channels'].update({cell_type: dict()})

        self.network_wide[self.name_to_key[neurotransmitter]]['ion_channels'][cell_type].update({section: ion_channels})

    def receptor_modulation(self, neurotransmitter, cell_type, receptor, modulation):

        if cell_type not in self.network_wide[self.name_to_key[neurotransmitter]]['receptors'].keys():
            self.network_wide[self.name_to_key[neurotransmitter]]['receptors'].update({cell_type: dict()})

        self.network_wide[self.name_to_key[neurotransmitter]]['receptors'][cell_type].update({receptor: modulation})

    def presynaptic_receptor_modulation(self, neurotransmitter, cell_type, receptor, modulation):

        if cell_type not in self.network_wide[self.name_to_key[neurotransmitter]]['presynaptic'].keys():

            self.network_wide[self.name_to_key[neurotransmitter]]['presynaptic'][cell_type].update({receptor: modulation})

    def plot_transient(self, neurotransmitter):

        import snudda.neuromodulation.modulation as modulation
        import matplotlib.pyplot as plt

        temp = copy.deepcopy(self.network_wide[self.name_to_key[neurotransmitter]])
        duration = np.arange(0, temp['duration'], self.dt)
        temp['parameters'].update({"time_step_array": duration})
        method = getattr(modulation, temp['method'])
        modulation_vector = method(temp['parameters'])

        plt.figure()
        plt.title(f" Transient for the modulation using {neurotransmitter} ")
        plt.plot(duration, modulation_vector)
        plt.ylabel("Modulation")
        plt.xlabel("Time (ms)")
        plt.show()

    def save(self, dir_path, name):

        cell_types = list()
        sections = ["axon", "dendrite", "soma"]
        modulation_types = ["ion_channels", "receptors", "presynaptic"]

        if not self.dt:
            raise ValueError(' Set time step for simulation')
        else:
            for neurotransmitter_type in self.network_wide.keys():
                self.network_wide[neurotransmitter_type].update({'dt': self.dt})

        # Add all the types of modulation, if they do not exist

        for m in modulation_types:

            for neurotransmitter_type in self.network_wide:
                if m not in self.network_wide[neurotransmitter_type]:
                    self.network_wide[neurotransmitter_type].update({m: dict()})

        # Add all cell types to all the types of modulation

        for neurotransmitter_type, modulation_information in self.network_wide.items():

            for m in modulation_information:
                if m in modulation_types:
                    for c in modulation_information[m]:
                        if c not in cell_types:
                            cell_types.append(c)

        for neurotransmitter_type, modulation_type in self.network_wide.items():

            if "ion_channels" in modulation_type:

                for cell_type in cell_types:

                    if cell_type in modulation_type["ion_channels"]:

                        for section in sections:
                            if section in modulation_type["ion_channels"][cell_type]:
                                pass
                            else:
                                modulation_type["ion_channels"][cell_type].update({section: list()})
                    else:
                        modulation_type["ion_channels"].update({cell_type: dict()})

                        for section in sections:
                            modulation_type["ion_channels"][cell_type].update({section: list()})

            if "receptors" in modulation_type:

                for cell_type in cell_types:

                    if cell_type in modulation_type["receptors"]:
                        pass
                    else:
                        modulation_type["receptors"].update({cell_type: dict()})

            if "presynaptic" in modulation_type:

                for cell_type in cell_types:

                    if cell_type in modulation_type["presynaptic"]:
                        pass
                    else:
                        modulation_type["presynaptic"].update({cell_type: dict()})

        temp = dict()
        temp.update({'type': self.type})
        temp.update({'description': self.network_wide})
        with open(os.path.join(dir_path, name), 'w') as out_file:
            json.dump(temp, out_file, indent=4, sort_keys=True)


if __name__ == "__main__":
    neurotransmitter = "dopamine"
    neurotransmitter_key = "DA"
    tstart = 300
    tonic = 0.2
    gmax_increase = 0.8
    tau = 500
    duration = 3000
    dt = 0.025
    nl = Neuromodulation()
    nl.set_timestep(dt=dt)
    name = "dopamine_modulation.json"
    dir_path = ""
    nl.set_modulation(neurotransmitter=neurotransmitter, neurotransmitter_key=neurotransmitter_key)
    nl.transient(neurotransmitter=neurotransmitter,
                 method="alpha_background",
                 duration=duration,
                 parameters={"tstart": tstart,
                             "tonic": tonic,
                             "gmax_increase": gmax_increase,
                             "tau": tau})

    nl.ion_channel_modulation(neurotransmitter=neurotransmitter,
                              cell_type="dSPN",
                              section="soma",
                              ion_channels=["kas_ms", "kaf_ms", "can_ms"])
    nl.ion_channel_modulation(neurotransmitter=neurotransmitter,
                              cell_type="dSPN",
                              section="dendrite",
                              ion_channels=["kas_ms", "kaf_ms"])

    nl.save(dir_path=dir_path, name=name)

    



# We want to investigate the distribution of glutamate and GABA synapses on the dendrites
#
# How many GABA synapses are there close to each glutamate synapse?
# Does it differ within the dendritic tree?
# Do we see clustering of synapses? -- This can be even more interesting to investigate after we have applied plasticity
#
#

import numpy as np
import os
from snudda import SnuddaLoad
import h5py

from snudda.neurons.morphology_data import MorphologyData


class PlotSynapseDistance:

    def __init__(self, network_path, network_file=None, input_file=None):

        self.network_path = network_path

        if network_file:
            self.network_file = network_file
        else:
            self.network_file = os.path.join(network_path, "network-synapses.hdf5")

        if input_file:
            self.input_file = input_file
        else:
            self.input_file = os.path.join(network_path, "input-spikes.hdf5")

        self.snudda_load = SnuddaLoad(network_file=network_file)
        self.input = h5py.File(self.input_file, "r")

    def gather_synapses(self, neuron_id):

        # Find the synapses from the connection matrix
        synapses = self.snudda_load.find_synapses(post_id=neuron_id)

        # Find input
        external_input_data = self.input["input"][str(neuron_id)]
        external_input = dict()

        for input_name in external_input_data.keys():
            external_input[input_name] = dict()
            external_input[input_name]["section_id"] = external_input_data[input_name].attrs["SectionID"].copy()
            external_input[input_name]["section_x"] = external_input_data[input_name].attrs["SectionX"].copy()

            # Do we also want to load the input spikes?  They are in neuron_input_data[input_name]["spikes"][()].copy()

        return synapses, external_input

    def load_morphology(self, neuron_id):

        morphology_path = self.snudda_load.data["neurons"][neuron_id]["morphology"]
        morphology_data = MorphologyData(swc_file=morphology_path)

        return morphology_data

    def get_synapse_neighbours(self, neuron_id, input_name,
                               soma_distance_range=None,
                               max_neighbour_distance=30e-6):

        # This function find the distances to the surrounding synapses from different external
        # input and synapse types

        # 1. Loop through all the sections (and note the distance to soma)
        # 2. Find the synapses belonging to input_name
        # 3. Find all parent and child comparments.
        # 4. Find all external input synapses (include of input_name) in the current section_id
        #    and in all the parent and child compartments
        # 5. Calculate distance to the input_name synapses in that section id.

        morphology_data = self.load_morphology(neuron_id=neuron_id)

        synapses, external_input = self.gather_synapses(neuron_id=neuron_id)

        neighbour_synapses = dict()

        for section in morphology_data.section_iterator(section_type=3):

            # Find the input locations
            idx = np.where(external_input[input_name]["section_id"] == section.section_id)[0]
            input_section_x = external_input[input_name]["section_x"][idx]
            input_soma_distance = section.soma_distance_at(input_section_x)
            input_soma_distance_all = input_soma_distance

            if soma_distance_range:
                keep_idx = np.where(np.logical_and(soma_distance_range[0] <= input_soma_distance,
                                                   input_soma_distance < soma_distance_range[1]))[0]
                input_section_x = input_section_x[keep_idx]
                input_soma_distance = input_soma_distance[keep_idx]

            # Now find all the neighbouring synapses of different types!
            for neighbour_input_name in external_input:
                neighbour_synapses[neighbour_input_name] = []

                # First we compare with synapses in the section itself
                neigh_dist = np.repeat(input_soma_distance.reshape(len(input_soma_distance_all, 1)),
                                       repeats=len(input_soma_distance_all),
                                       axis=1) \
                            - np.repeat(input_soma_distance_all.reshape(1, len(input_soma_distance)),
                                        repeats=len(input_soma_distance),
                                        axis=0)

                # !!! TODO: Work in progress

                # Then in the parent section

                # The in the children sections

            neighbour_section_id = [section.section_id, section.parent_section_id] + list(section.child_section_id)

            neighbour_idx = np.where(external_input[input_name]["section_id"] in neighbour_section_id)[0]
            neighbour_section_x = external_input[input_name]["section_x"][neighbour_idx]
            neighbour_soma_distance = se

            parent_section = morphology_data.section_data[3][section.parent_section_id]
            child_sections = [morphology_data.section_data[3][s_id] for s_id in section.child_section_id]



        pass



bl_info = {
    "name": "SWC format",
    "author": "Martin Pyka",
    "version": (1, 0, 0),
    "blender": (2, 74, 0),
    "location": "File > Import > SWC",
    "description": "Import SWC files",
    "license": "GPL v2",
    "category": "Import"}

__author__ = bl_info['author']
__license__ = bl_info['license']
__version__ = ".".join([str(s) for s in bl_info['version']])

import bpy
from . import operator_swc_import

def register():
    #bpy.utils.register_module(__name__)
    operator_swc_import.register()
    
def unregister():
    #bpy.utils.unregister_module(__name__)
    operator_swc_import.unregister()

if __name__ == "__main__":
    register()




# This uses ConvHurt to convert the network created by Network_connect.py

#
# TODO:
# - Add support for gap junctions
# - Check what axon and dendrite propagation speeds should be
#

import json
import os
import sys
from collections import OrderedDict
from shutil import copyfile
from glob import glob

import h5py
import numpy as np

from snudda.utils import snudda_parse_path
from snudda.utils.conv_hurt import ConvHurt
from snudda import SnuddaLoad


class ExportSonata:

    def __init__(self, network_path=None, network_file=None, input_file=None, out_dir=None, debug_flag=True, target_simulator="NEST"):

        self.debug = debug_flag

        if network_path is not None:
            self.network_path = network_path
        elif network_file is not None:
            self.network_path = os.path.dirname(network_file)
        else:
            raise ValueError("You must specify network_path or network_file")

        if network_file is None:
            network_file = os.path.join(network_path, "network-synapses.hdf5")

        if out_dir is not None:
            self.out_dir = out_dir
        else:
            self.out_dir = os.path.join(network_path, "SONATA")

        self.target_simulator = target_simulator

        # Read the data
        self.snudda_load = SnuddaLoad(network_file)
        self.network_file = self.snudda_load.network_file

        if input_file is not None:
            self.input_file = input_file
        else:
            input_file_candidate = os.path.join(self.network_path, "input-spikes.hdf5")

            if os.path.isfile(input_file_candidate):
                self.input_file = input_file_candidate
            else:
                self.input_file = None

        self.network_config = json.loads(self.snudda_load.data["config"])

        if self.input_file:
            print(f"Using input file: {self.input_file}")
            has_input = True
        else:
            print("No input file specified, and default file not found.")
            has_input = False

        # This contains data for converting to Neurodamus secID and secX
        self.morph_cache = dict([])

        self.morph_location_lookup = dict([])
        self.hoc_location_lookup = dict([])

        # Write the data in new format using the ConvHurt module
        # TODO: We need to read structure names from the network-config.json
        structure_names = [x for x in self.network_config["regions"]]
        ch = ConvHurt(simulation_structures=structure_names,
                      base_dir=self.out_dir, has_input=has_input)

        self.copy_morphologies()
        self.copy_mechanisms()
        self.copy_hoc()

        # Convert our rotation matrices to vx,vy,vz rotation angles
        # Neurodamus rotation order is first Z then Y and X.
        (vx, vy, vz) = self.rot_angles_zyx(self.snudda_load)

        # We need to convert the named neuron types to numbers
        node_type_id, node_type_id_lookup = self.allocate_node_type_id()
        # node_group_id, group_idx, node_group_lookup = self.allocate_node_groups()
        node_group_id, group_idx, node_group_lookup, neuron_id_remap = self.allocate_groups_and_remap_nodeid()

        volume_id_list = [x["volume_id"] for x in self.snudda_load.data["neurons"]]
        volume_list = set(volume_id_list)

        # node_name_list = [x["name"] for x in self.snudda_load.data["neurons"]]
        node_name_list = [f"{n['name']}_{n['morphology_key']}_{n['parameter_key']}_{n['modulation_key']}"
                          for n in self.snudda_load.data["neurons"]]

        # Edge data is stored in a HDF5 file and a CSV file
        edge_type_lookup = dict()
        for (pre_type, post_type), con_data in self.snudda_load.data["connectivity_distributions"].items():
            for con_type, con_type_data in con_data.items():
                if con_type == "gap_junction":
                    # TODO: How do we write gap junctions to SONATA?
                    continue
                else:
                    edge_type_id = con_type_data["channel_model_id"]

                    if self.target_simulator == "NEST":
                        if "nest_model_template" in con_type_data["channel_parameters"]:
                            edge_model = con_type_data["channel_parameters"]["nest_model_template"]
                            if "nest_dynamic_params" not in con_type_data["channel_parameters"]:
                                raise KeyError("If nest_model_template is specified, nest_dynamic_params must be specified also")
                            dynamic_params = con_type_data["channel_parameters"]["nest_dynamic_params"]
                        else:
                            edge_model = "static_synapse"
                            if con_type == "GABA":
                                dynamic_params = "inhibitory.json"
                            else:
                                dynamic_params = "excitatory.json"
                    else:
                        edge_model = con_type_data["channel_parameters"]["mod_file"]

                edge_type_lookup[pre_type, post_type, con_type] = (edge_type_id, edge_model, f"{pre_type}_{post_type}", dynamic_params)

        node_id_remap = np.full(shape=(self.snudda_load.data["num_neurons"],), fill_value=-1, dtype=int)
        input_list = []

        for volume_name in volume_list:

            node_file = f"{volume_name}_nodes.hdf5"

            for nt in self.snudda_load.get_neuron_types(return_set=True):
                nt_idx = np.where([x["volume_id"] == volume_name and x["type"] == nt
                                   for x in self.snudda_load.data["neurons"]])

                # Node data is stored in a HDF5 file and a CSV file (CONVERT TO micrometers)
                node_data = OrderedDict([("x", self.snudda_load.data["neuron_positions"][nt_idx, 0].flatten() * 1e6),
                                         ("y", self.snudda_load.data["neuron_positions"][nt_idx, 1].flatten() * 1e6),
                                         ("z", self.snudda_load.data["neuron_positions"][nt_idx, 2].flatten() * 1e6),
                                         ("rotation_angle_zaxis", vz[nt_idx].flatten()),
                                         ("rotation_angle_yaxis", vy[nt_idx].flatten()),
                                         ("rotation_angle_xaxis", vx[nt_idx].flatten())])

                node_id_list = np.array([x["neuron_id"] for x in self.snudda_load.data["neurons"]
                                         if x["volume_id"] == volume_name and x["type"] == nt], dtype=int)

                assert (nt_idx == node_id_list).all()

                new_node_id = np.arange(0, len(node_id_list))

                assert (node_id_remap[node_id_list] == -1).all(), f"Node id already remapped!"
                node_id_remap[node_id_list] = new_node_id

                # Population name should be name of neuron type
                node_file = ch.write_nodes(node_file=node_file,
                                           population_name=nt,
                                           data=node_data,
                                           node_id=node_id_remap[node_id_list],
                                           node_type_id=node_type_id[nt_idx],
                                           node_group_id=node_group_id[nt_idx],
                                           node_group_index=group_idx[nt_idx],
                                           close_file=False)

            node_file.close()
            # Done writing hdf5 file, time to create csv file

            assert (node_id_remap >= 0).all(), f"Not all node id remapped."

            v_idx = np.where([v == volume_name for v in volume_id_list])[0]
            csv_node_names = sorted(list(set([node_name_list[x] for x in v_idx])))
            csv_node_type_id = [node_type_id_lookup[n] for n in csv_node_names]
            csv_node_location = [volume_name for n in csv_node_names]

            (model_template_list, model_type_list, morphology_list, dynamic_params) \
                = self.get_node_templates(csv_node_names)

            # Gather all the NEST models
            dynamic_params = self.copy_and_rewrite_dynamics_params_files(dynamics_params_list=dynamic_params)

            node_data_csv = OrderedDict([("model_type", model_type_list),
                                         ("model_template", model_template_list),
                                         ("location", csv_node_location),
                                         ("morphology", morphology_list),
                                         ("model_name", csv_node_names)])

            if dynamic_params is not None and len(dynamic_params) > 0:
                node_data_csv["dynamics_params"] = dynamic_params

            ch.write_node_csv(node_csv_file=f"{volume_name}_node_types.csv",
                              node_type_id=csv_node_type_id,
                              data=node_data_csv)

            edge_population_lookup, edge_population_id_lookup = self.setup_edge_population(node_group_lookup)

            population_rows, edge_type_id, source_gid, target_gid, edge_data = \
                self.setup_edge_info(edge_population_lookup, node_group_id, group_idx)

            ch.write_edges(edge_file=f"{volume_name}_edges.hdf5",
                           population_rows=population_rows,
                           edge_type_id=edge_type_id,
                           source_id=source_gid,
                           target_id=target_gid,
                           data=edge_data)

            edge_type_id = [x[0] for x in edge_type_lookup.values()]

            edge_data_csv = {"model_template": [x[1] for x in edge_type_lookup.values()],
                             "population": [x[2] for x in edge_type_lookup.values()],
                             "dynamic_params": [x[3] for x in edge_type_lookup.values()]}

            ch.write_edges_csv(edge_csv_file=f"{volume_name}_edge_types.csv",
                               edge_type_id=edge_type_id,
                               data=edge_data_csv)

            if self.input_file:

                (sonata_input_hdf5,
                 sonata_virtual_node_hdf5, sonata_virtual_node_csv,
                 sonata_virtual_edges_hdf5, sonata_virtual_edges_csv) = \
                    self.write_input(node_id_remap=node_id_remap,
                                     input_hdf5=self.input_file,
                                     conv_hurt=ch, volume_name=volume_name)

                input_list.append((volume_name, sonata_input_hdf5,
                                   sonata_virtual_node_hdf5, sonata_virtual_node_csv,
                                   sonata_virtual_edges_hdf5, sonata_virtual_edges_csv))

        # !!! TODO: We need to pass the new input files, node and edge files to the config
        self.write_simulation_config(input_list=input_list)

        print(f"SONATA files exported to {self.out_dir}")

    ############################################################################

    def get_all_input_types(self, input_hdf5, neuron_type=None, volume_name=None):

        input_types = set()

        if type(input_hdf5) != h5py._hl.files.File:
            input_hdf5 = h5py.File(input_hdf5, "r")

        neuron_id_list = self.snudda_load.get_neuron_id_of_type(neuron_type=neuron_type, volume=volume_name)

        for neuron_id in neuron_id_list:
            neuron_input_types = set(input_hdf5[f"input/{neuron_id}"].keys())

            input_types += neuron_input_types

        return input_types

    def write_input(self,
                    input_hdf5,
                    conv_hurt,
                    node_id_remap,
                    volume_name=None):

        """
            Args:
                input_hdf5 : File to read data from
                sonata_input_hdf5 : File to write data to
                conv_hurt :
                node_id_remap : Numpy array mapping neuron_id to gid (which is population specific)
                volume_name (str) : Name of volume (default: None, assumes neuron type only in one volume)
        """

        # If there are N neurons in the volume, and they receive cortical and thalamic input
        #
        # Then...
        # Virtual neuron 0-(N-1) provide cortical input to neurons 0-(N-1)
        # Virtual neuron N-(2*N-1) provide thalamic input to neurons 0-(N-1)
        #

        if type(input_hdf5) != h5py._hl.files.File:
            input_hdf5 = h5py.File(input_hdf5, "r")

        sonata_input_hdf5 = f"inputs/input_{volume_name}.hdf5"

        neuron_id_list = self.snudda_load.get_neuron_id_of_type(volume=volume_name, neuron_type=None)
        input_data = []  # [(neuron_type, neuron_node_id, input_type, spikes, weight, virtual_node_id), (np, nid, it, s, w, vid), ...]

        virtual_node_ctr = 0

        for neuron_id in neuron_id_list:

            for input_type in input_hdf5[f"input/{neuron_id}"].keys():
                neuron_spikes = []

                spike_group = input_hdf5[f"input/{neuron_id}/{input_type}/spikes"]
                n_spikes = spike_group.attrs["num_spikes"]

                for idx, ns in enumerate(n_spikes):
                    neuron_spikes.append(spike_group[idx, :ns])

                input_spikes = np.array(sorted(np.concatenate(neuron_spikes)))

                neuron_type = self.snudda_load.data["neurons"][neuron_id]["type"]

                weight = input_hdf5[f"input/{neuron_id}/{input_type}"].attrs["conductance"] * 1e6  # microsiemens for NEST
                if "GABA" in input_hdf5[f"input/{neuron_id}/{input_type}"].attrs["mod_file"].upper():
                    weight *= -1

                # [(neuron_type, neuron_node_id, input_type, spikes, weight, virtual_node_id), (np, nid, it, s, w, vid), ...]
                input_data.append((neuron_type, node_id_remap[neuron_id], input_type, input_spikes, weight, virtual_node_ctr))
                virtual_node_ctr += 1

        # Next we need to write the input spikes to file
        all_spikes = []
        all_gid = []

        for _, _, _, spikes, _, gid in input_data:
            all_spikes.append(spikes)
            all_gid.append(np.full(shape=spikes.shape, fill_value=gid, dtype=int))

        spikes = np.concatenate(all_spikes)
        gids = np.concatenate(all_gid)

        if len(spikes) == 0:
            return None

        # We need to make sure the spikes are sorted by gid, then by time
        idx = np.lexsort((spikes, gids))

        f_name = conv_hurt.write_input(spike_file_name=sonata_input_hdf5,
                                       spike_times=spikes[idx],
                                       gids=gids[idx])

        input_hdf5.close()

        # We also need to create the virtual nodes, and edges to connect them

        virtual_node_id = np.array([x[5] for x in input_data], dtype=int)
        neuron_type = [x[0] for x in input_data]
        node_id = np.array([x[1] for x in input_data], dtype=int)
        weight = np.array([x[4] for x in input_data])

        sonata_virtual_node_hdf5, sonata_virtual_node_csv, sonata_virtual_edges_hdf5, sonata_virtual_edges_csv = \
            self.add_virtual_input(volume_name=volume_name, virtual_node_id=virtual_node_id,
                                   neuron_type=neuron_type, node_id=node_id, weight=weight, conv_hurt=conv_hurt)

        return (sonata_input_hdf5,
                sonata_virtual_node_hdf5, sonata_virtual_node_csv,
                sonata_virtual_edges_hdf5, sonata_virtual_edges_csv)

    ############################################################################

    def rot_mat_to_angles_xyz(self, rot_mat):
        # http://nghiaho.com/?page_id=846

        # This is for Rz*Ry*Rx
        vx = np.arctan2(rot_mat[2, 1], rot_mat[2, 2])
        vy = np.arctan2(-rot_mat[2, 0], np.sqrt(rot_mat[2, 1] ** 2 + rot_mat[2, 2] ** 2))
        vz = np.arctan2(rot_mat[1, 0], rot_mat[0, 0])

        # !! We need to do this for Rx*Ry*Rz
        assert False

        return vx, vy, vz

    ############################################################################

    def rot_mat_to_angles_zyx(self, rotMat):
        vz = np.arctan2(-rotMat[0, 1], rotMat[0, 0])
        vy = np.arcsin(rotMat[0, 2])
        vx = np.arctan2(-rotMat[1, 2], rotMat[2, 2])

        return vx, vy, vz

    ############################################################################

    def angles_to_rot_mat(self, v):
        (vx, vy, vz) = v
        RX = np.array([[1., 0., 0.],
                       [0., np.cos(vx), -np.sin(vx)],
                       [0., np.sin(vx), np.cos(vx)]])

        RY = np.array([[np.cos(vy), 0., np.sin(vy)],
                       [0., 1., 0.],
                       [-np.sin(vy), 0, np.cos(vy)]])

        RZ = np.array([[np.cos(vz), -np.sin(vz), 0.],
                       [np.sin(vz), np.cos(vz), 0.],
                       [0., 0., 1.]])

        return np.matmul(RX, np.matmul(RY, RZ))
        # return np.matmul(np.matmul(RZ,RY),RX)

    ############################################################################

    def rot_angles_zyx(self, nl):

        x = np.zeros(len(nl.data["neurons"]))
        y = np.zeros(len(nl.data["neurons"]))
        z = np.zeros(len(nl.data["neurons"]))

        for idx in range(0, len(nl.data["neurons"])):
            ang = self.rot_mat_to_angles_zyx(nl.data["neurons"][idx]["rotation"])

            # Just verify that we calculated the angles right and can get old
            # rotation matrix back
            assert (np.sum(np.abs(self.angles_to_rot_mat(ang) - nl.data["neurons"][idx]["rotation"])) < 1e-9)
            # Save angles
            x[idx] = ang[0]
            y[idx] = ang[1]
            z[idx] = ang[2]

        return x, y, z

    ############################################################################

    def allocate_node_type_id(self):

        # Since each neuron directory can have multiple morphology/parameter sets
        # we need to use the name + morphology_key + parameter_key + modulation_key

        node_type_id_lookup = OrderedDict([])
        next_node_type_id = 0

        node_names = [f"{n['name']}_{n['morphology_key']}_{n['parameter_key']}_{n['modulation_key']}"
                      for n in self.snudda_load.data["neurons"]]

        node_names_sorted = sorted(node_names)

        for n in node_names_sorted:
            if n not in node_type_id_lookup:
                node_type_id_lookup[n] = next_node_type_id
                next_node_type_id += 1

        node_type_id = np.array([node_type_id_lookup[x] for x in node_names], dtype=int)

        return node_type_id, node_type_id_lookup

    ############################################################################

    # Comment from 2018, hope they support all now (guess we will find out):
    #  --> RT neuron only supports one group, so put everything in the same group

    # Snudda neuron type corresponds to SONATA NodeGroup
    # SONATA NodeType is {Snudda neuron name}_{morphology_key}_{parameter_key}_{modulation_key}

    def allocate_node_groups_OLD(self):

        # node_group_id -- tells which group_id each neuron belongs to
        # group_idx -- tells the index with the group that the neuron has

        node_group_id = np.zeros(len(self.snudda_load.data["neurons"]), dtype=int)
        group_idx = np.zeros(len(self.snudda_load.data["neurons"]), dtype=int)
        group_lookup = dict([])

        neuron_types = [n["type"] for n in self.snudda_load.data["neurons"]]

        next_group_idx = dict()
        next_group = 0

        for nt in neuron_types:
            if nt not in group_lookup:
                group_lookup[nt] = next_group
                next_group += 1
                next_group_idx[nt] = 0

        for idx, nt in enumerate(neuron_types):
            node_group_id[idx] = group_lookup[nt]
            group_idx[idx] = next_group_idx[nt]
            next_group_idx[nt] += 1

        return node_group_id, group_idx, group_lookup

    def allocate_groups_and_remap_nodeid(self):

        n_neurons = self.snudda_load.data["num_neurons"]
        volume_list = set([x["volume_id"] for x in self.snudda_load.data["neurons"]])
        neuron_types = self.snudda_load.get_neuron_types(return_set=True)

        node_group_id = np.zeros(shape=(n_neurons, ), dtype=int)
        group_idx = np.full(shape=(n_neurons, ), fill_value=-1, dtype=int)
        neuron_id_remap = np.full(shape=(n_neurons, ), fill_value=-1, dtype=int)

        next_group = 0
        group_lookup = dict()

        for vol in volume_list:
            for nt in neuron_types:
                idx = self.snudda_load.get_neuron_id_of_type(neuron_type=nt, volume=vol)

                group_lookup[vol, nt] = next_group
                node_group_id[idx] = next_group
                next_group += 1

                assert (neuron_id_remap[idx] == -1).all(), f"Neuron id already remapped!"
                neuron_id_remap[idx] = np.arange(0, len(idx))
                group_idx[idx] = np.arange(0, len(idx))

                # Originally we did not remap neuron id, but it seems that SONATA and NEST requires
                # renumbering of the neuron_id to be consequent and starting from 0 for each group

        assert (neuron_id_remap >= 0).all(), f"Not all neuron_id remapped!"

        return node_group_id, group_idx, group_lookup, neuron_id_remap

    def remap_nodes(self):

        n_neurons = self.snudda_load.data["num_neurons"]
        neuron_id_remap = np.full(shape=(n_neurons,), fill_value=-1)

        neuron_types = self.snudda_load.get_neuron_types()
        for nt in neuron_types:
            old_id = self.snudda_load.get_neuron_id_of_type(neuron_type=nt)
            new_id = np.arange(0, len(old_id))

            assert (neuron_id_remap[old_id] == -1).all(), f"old_id already allocated!"
            neuron_id_remap[old_id] = new_id

        assert (neuron_id_remap >= 0).all(), f"Not all neuron_id remapped"

        return neuron_id_remap

    ############################################################################

    def get_node_templates(self, csv_node_name):

        template_dict = dict([])
        model_type_dict = dict([])
        morphology_dict = dict([])
        missing_list = []

        dynamics_params_list = []  # Used by NEST

        # First create a lookup
        for n in self.snudda_load.data["neurons"]:
            hoc_file = n["hoc"]
            name = f"{n['name']}_{n['morphology_key']}_{n['parameter_key']}_{n['modulation_key']}"

            # It seems that RT neuron requires the filename without .swc at the end
            # Also, they prepend morphology path, but only allows one directory for all morphologies
            # morph_file = os.path.splitext(os.path.basename(n["morphology"]))[0]
            morph_file = os.path.basename(n["morphology"]).replace(".swc", "")

            if hoc_file in self.hoc_location_lookup:
                # We need to change from old hoc location to the SONATA hoc location
                hoc_str = f"hoc:{os.path.basename(self.hoc_location_lookup[hoc_file])}"
            elif n["virtual_neuron"]:
                hoc_str = "virtual"
            else:
                hoc_str = "NULL"
                if hoc_file not in missing_list:
                    # Only write error ones per file
                    print(f"Missing hoc template: {hoc_file}")
                    missing_list.append(hoc_file)

            if name not in morphology_dict:
                morphology_dict[name] = morph_file
            else:
                assert os.path.basename(morphology_dict[name]) == os.path.basename(morph_file), \
                    f"Morphology mismatch for {name}: {morphology_dict[name]} vs {morph_file}"

            if name not in template_dict:
                if self.target_simulator == "NEST":
                    model_type_dict[name] = "point_neuron"
                    dynamics_params_list.append(self.get_dynamics_params(n))
                    template_dict[name] = "aeif_cond_exp"
                else:
                    model_type_dict[name] = "biophysical"
                    template_dict[name] = hoc_str

            elif self.target_simulator != "NEST":

                assert template_dict[name] == hoc_str, \
                    f"All files named {name} do not share same hoc file: {template_dict[name]} and {hoc_str}"

        # Need to put them in the order in csvNodeName

        template_list = [template_dict[x] for x in csv_node_name]
        model_type_list = [model_type_dict[x] for x in csv_node_name]
        morph_list = [morphology_dict[x] for x in csv_node_name]

        return template_list, model_type_list, morph_list, dynamics_params_list

    ############################################################################

    def copy_and_rewrite_dynamics_params_files(self, dynamics_params_list):

        dest_path = os.path.join(self.out_dir, "components", "point_neuron_dynamics")

        new_dynamics_params_list = []
        copied_files = []

        for file_path in dynamics_params_list:

            if file_path is None:
                new_dynamics_params_list.append("MISSING-FILE-HERE.json")
                continue

            if os.path.basename(file_path) == "dynamics_params.json":
                dir_name = os.path.basename(os.path.dirname(file_path))
                new_file = os.path.join(dest_path, f"{dir_name}_dynamics_params.json")
            else:
                new_file = os.path.join(dest_path, os.path.basename(file_path))

            if new_file not in copied_files:
                copyfile(file_path, new_file)
                copied_files.append(new_file)

            new_dynamics_params_list.append(os.path.basename(new_file))

        return new_dynamics_params_list

    def setup_edge_population(self, node_group_lookup):

        edge_population_lookup = dict()
        edge_population_id_lookup = dict()
        next_id = 1

        for (pre_volume, pre_node_type), pre_node_group in node_group_lookup.items():
            for (post_volume, post_node_type), post_node_group in node_group_lookup.items():

                if (pre_node_group, post_node_group) in edge_population_lookup:
                    raise KeyError(f"{pre_node_group} to {post_node_group} connections occur in multiple volumes, "
                                   f"please rename so neuron with specific name only occur in one volume")

                edge_population_lookup[pre_node_group, post_node_group] = f"{pre_node_type}_{post_node_type}"
                edge_population_id_lookup[pre_node_group, post_node_group] = next_id
                next_id += 1

        return edge_population_lookup, edge_population_id_lookup

    ############################################################################

    def get_nest_synapse_models(self):

        synapse_model_lookup = dict()

        for synapse_models in self.snudda_load.data["connectivity_distributions"].values():
            for synapse_type, synapse_model in synapse_models.items():
                synapse_type_id = synapse_model["channel_model_id"]

                if synapse_type == "GABA":
                    synapse_model_lookup[synapse_type_id] = dict()
                    synapse_model_lookup[synapse_type_id]["soma"] = "inhibitory_soma.json"
                    synapse_model_lookup[synapse_type_id]["proximal"] = "inhibitory_proximal.json"
                    synapse_model_lookup[synapse_type_id]["distal"] = "inhibitory_distal.json"
                else:
                    # Assume excitatory
                    synapse_model_lookup[synapse_type_id] = dict()
                    synapse_model_lookup[synapse_type_id]["soma"] = "excitatory_soma.json"
                    synapse_model_lookup[synapse_type_id]["proximal"] = "excitatory_proximal.json"
                    synapse_model_lookup[synapse_type_id]["distal"] = "excitatory_distal.json"

        return synapse_model_lookup

    def get_nest_synapse_sign_lookup(self):

        synapse_sign_lookup = dict()

        for synapse_models in self.snudda_load.data["connectivity_distributions"].values():
            for synapse_type, synapse_model in synapse_models.items():
                synapse_type_id = synapse_model["channel_model_id"]

                if synapse_type.upper() == "GABA":
                    synapse_sign_lookup[synapse_type_id] = -1.0
                else:
                    synapse_sign_lookup[synapse_type_id] = 1.0

        return synapse_sign_lookup

    ############################################################################

    # This code sets up the info about edges

    def setup_edge_info(self, edge_population_lookup, node_group_id, group_idx):

        n_synapses = self.snudda_load.data["synapses"].shape[0]
        edge_type_id = np.zeros(n_synapses, dtype=int)
        source_gid = np.zeros(n_synapses, dtype=int)
        target_gid = np.zeros(n_synapses, dtype=int)

        population_rows = dict()  # For each population name, list all the synapse rows

        sec_id = np.zeros(n_synapses, dtype=int)
        sec_x = np.zeros(n_synapses)
        syn_weight = np.zeros(n_synapses)
        delay = np.zeros(n_synapses)

        # Check if these speeds are reasonable?
        axon_speed = 25.0  # 25m/s
        dend_speed = 1.0

        # columns in nl.data["synapses"]
        # 0: sourceCellID, 1: sourceComp, 2: destCellID, 3: destComp,
        # 4: locType, 5: synapseType, 6: somaDistDend 7:somaDistAxon
        # somaDist is an int, representing micrometers

        synapse_sign_lookup = self.get_nest_synapse_sign_lookup()

        for i_syn, syn_row in enumerate(self.snudda_load.data["synapses"]):
            source_gid[i_syn] = group_idx[syn_row[0]]
            target_gid[i_syn] = group_idx[syn_row[1]]

            source_node_group = node_group_id[syn_row[0]]
            target_node_group = node_group_id[syn_row[1]]

            population_group = edge_population_lookup[source_node_group, target_node_group]

            if population_group in population_rows:
                population_rows[population_group].append(i_syn)
            else:
                population_rows[population_group] = [i_syn]

            sec_id[i_syn] = syn_row[9]
            sec_x[i_syn] = syn_row[10] / 1000.0
            synapse_type = syn_row[6]
            synapse_conductance = syn_row[11] * 1e-3  # pS --> micro simens

            # We need to set the weight to negative for inhibitory synapses it seems
            # hence the synapse_sign_lookup (returns +1 or -1)
            syn_weight[i_syn] = synapse_conductance * synapse_sign_lookup[synapse_type]

            pre_type = self.snudda_load.data["neurons"][source_gid[i_syn]]["type"]
            post_type = self.snudda_load.data["neurons"][target_gid[i_syn]]["type"]

            edge_type_id[i_syn] = synapse_type

            dend_dist = syn_row[6] * 1e-6
            axon_dist = syn_row[7] * 1e-6
            delay[i_syn] = axon_dist / axon_speed * 1e3 + 1  # Delay in ms and not SI units :-(

        edge_data = OrderedDict([("afferent_section_id", sec_id),
                                 ("afferent_section_pos", sec_x),
                                 ("syn_weight", syn_weight),
                                 ("delay", delay)])

        for pop_name in population_rows.keys():
            population_rows[pop_name] = np.array(population_rows[pop_name])

        return population_rows, edge_type_id, source_gid, target_gid, edge_data

    ############################################################################

    # Convert synapse coordinates (in SWC reference frame) to SectionID
    # and section_X (fractional distance along dendritic section)

    def conv_synapse_coord_to_section(self, cell_gid, orig_synapse_coord, network_info):

        assert False, "Depricated function: convSynapseCoordToSection"

        if self.debug:
            print("Converting coordinate to sections")

        cell_id = network_info.data["neurons"][cell_gid]["name"]

        # Get swc filename from cellID
        swc_file = network_info.config[cell_id]["morphology"]

        # Check if cell morphology is loaded in neuron cache
        # Columns in morph: 0:swcID, 1:x, 2:y, 3:z, 4:secID, 5:secX, 6:type
        if swc_file in self.morph_cache:
            morph = self.morph_cache[swc_file]
        else:
            morph = self.load_morph(swc_file)
            self.morph_cache[swc_file] = morph

            # VERIFICATION
            if False:
                self.plot_section_data(morph, swc_file)
                import pdb
                pdb.set_trace()

        # Find points matching the coords
        xyz = morph[:, 1:4]
        d = np.sum((xyz - orig_synapse_coord) ** 2, axis=-1)

        closest_idx = np.argsort(d)[0]
        sec_id = int(morph[closest_idx, 4])
        sec_x = morph[closest_idx, 5]

        # Check that compartment that matched is soma or dendrite
        try:
            assert morph[closest_idx, 6] in [1, 3, 4], \
                "Synapse location should be soma or dendrite"
        except:
            import pdb
            pdb.set_trace()

        if self.debug:
            print("match dmin=" + str(d[closest_idx] ** 0.5))

        # import pdb
        # pdb.set_trace()

        return sec_id, sec_x

    ############################################################################

    def get_dynamics_params(self, neuron):

        if self.target_simulator == "NEST":
            # If the dynamics_params.json file exist in the neuron path, then use that
            neuron_path = os.path.relpath(snudda_parse_path(neuron["neuron_path"], self.snudda_load.data["snudda_data"]))
            dynamics_params_path = os.path.join(neuron_path, "dynamics_params.json")

            if os.path.isfile(dynamics_params_path):
                return dynamics_params_path

            # Otherwise, if there is a default model in the nest/models directory, use that
            neuron_type = neuron["type"]
            alt_dynamics_path = os.path.join(self.snudda_load.data["snudda_data"],
                                             "nest", "models", f"{neuron_type}.json")
            if os.path.isfile(alt_dynamics_path):

                print(f"Missing {dynamics_params_path} file, using {alt_dynamics_path}")

                return alt_dynamics_path

            return None
        else:
            return None

    ############################################################################

    def convert_coordinates(self, morphology):

        # Loop through entire SWC morphology, calculate SEC_ID, SEC_X
        # for each vertex.

        print("TEST TEST TEST")
        import pdb
        pdb.set_trace()

    ############################################################################

    # This code builds the secID and secX from the SWC file

    def load_morph(self, swc_file, skip_axon=True):

        with open(swc_file, 'r') as f:
            lines = f.readlines()

        comp_type = {1: "soma", 2: "axon", 3: "dend", 4: "apic"}

        swc_vals = np.zeros(shape=(len(lines), 7))

        n_comps = 0
        for ss in lines:
            if ss[0] != '#':
                swc_vals[n_comps, :] = [float(s) for s in ss.split()]
                n_comps = n_comps + 1

        # Subtract 1 from ID and parentID, so we get easier indexing
        swc_vals[:, 0] -= 1
        swc_vals[:, 6] -= 1

        swc_vals[:, 2:5] *= 1e-6  # Convert to meter

        # Columns in vertex matrix:
        # 0:ID, 1:parentID, 2:x,3:y,4:z, 5:somaDist, 6:type, 7:childCount,
        # 8:nodeParent, 9:segLen, 10:seg_ID, 11:seg_X
        #
        # If childCount != 1 then vertex is a segment node
        vertex = np.zeros(shape=(n_comps, 12))

        for i in range(0, n_comps):
            # Check that the IDs are in order, and no missing values
            assert (swc_vals[i, 0] == i)

            # Check that parent already defined
            assert (swc_vals[i, 6] < swc_vals[i, 0])

            vertex[i, 0] = swc_vals[i, 0]  # swcID
            vertex[i, 2:5] = swc_vals[i, 2:5]  # x,y,z
            vertex[i, 6] = swc_vals[i, 1]  # type

            if swc_vals[i, 1] == 1:
                assert (i == 0)  # Soma should be first compartment

                vertex[i, 1] = -1  # parent_swcID
                vertex[i, 5] = 0  # soma dist
                vertex[i, 7] = 100  # Set child count > 1 for soma
                # to make sure it is a node

            elif swc_vals[i, 1] in [2, 3, 4]:
                # axon or dend
                parent_id = int(swc_vals[i, 6])  # parentID
                vertex[i, 1] = parent_id
                parent_coords = vertex[parent_id, 2:5]
                parent_dist = vertex[parent_id, 5]
                vertex[i, 5] = parent_dist + sum((vertex[i, 2:5] - parent_coords) ** 2) ** 0.5  # Soma dist

                # if(parentID == 1):
                #  # Soma as parent, set artificially high child count for compartment
                #  # so it is considered a node edge
                #  vertex[i,7] = 100
                # else:
                vertex[parent_id, 7] += 1  # Child count, nodes have != 1, i.e. 0 or 2,3,..

        if skip_axon:
            # Find all branch points and end points (excluding axons)
            node_idx = np.where(np.logical_and(vertex[:, 7] != 1, vertex[:, 1] != 2))[0]
        else:
            # Find all branch points and end points
            node_idx = np.where(vertex[:, 7] != 1)[0]

        # Find out how branch points are connected (set node parent)
        # OBS, parent is previous point in SWC file a point is connected to
        # For segments, node and node parent are the two end points
        for n_idx in node_idx:
            parent_id = int(vertex[n_idx, 1])
            vertex[n_idx, 8] = parent_id

            while vertex[parent_id, 7] == 1:
                # While the parent node has 1 child, and it is not soma
                parent_id = int(vertex[parent_id, 1])

                if vertex[parent_id, 6] == 1:
                    # This is a soma, dont use it as parent
                    break

                assert (n_idx == 0 or parent_id >= 0)
                vertex[n_idx, 8] = parent_id  # node parent

        # Number the segments, start with soma which has ID 0
        assert ((np.diff(node_idx) > 0).all())  # Check they are in right order

        for sec_id, n_idx in enumerate(node_idx):
            vertex[n_idx, 10] = sec_id

        # For each connected pair of branch points (dont forget soma)
        # calculate the total length of the segment (also note if soma,axon or dend)

        for n_idx in node_idx:

            if n_idx == 0:
                vertex[n_idx, 9] = 1  # segLen (to avoid divide by zero for soma)
            else:
                # Difference in soma dist between node and node parent
                parent_id = int(vertex[n_idx, 8])
                vertex[n_idx, 9] = vertex[n_idx, 5] - vertex[parent_id, 5]  # segLen

        # For each vertex in a segment, calculate the seg_X (value 0 to 1)
        # Save seg_ID and seg_X for each vertex, its coordinates, and S/A/D type

        for n_idx in node_idx:

            if n_idx == 0:
                vertex[n_idx, 11] = 0.5
            # elif(vertex[nIdx,1] == 0):
            #  # Soma is the parent, special case (Neuron has no compartment
            #  # linking soma centre to first axon/dendirte node)
            #  vertex[nIdx,11] = 0
            else:
                vertex[n_idx, 11] = 1  # secX
                end_soma_dist = vertex[n_idx, 5]
                seg_len = vertex[n_idx, 9]

                parent_id = int(vertex[n_idx, 1])

                while vertex[parent_id, 7] == 1:  # While the parent node has 1 child
                    # secID
                    vertex[parent_id, 10] = vertex[n_idx, 10]

                    # secX
                    vertex[parent_id, 11] = 1 - (end_soma_dist - vertex[parent_id, 5]) / seg_len
                    parent_id = int(vertex[parent_id, 1])

                    assert (parent_id >= 0)

        # swcID, x,y,z coord, secID, secX, type
        if skip_axon:
            non_axon_idx = np.where(vertex[:, 6] != 2)[0]
            return vertex[:, [0, 2, 3, 4, 10, 11, 6]][non_axon_idx, :]
        else:
            return vertex[:, [0, 2, 3, 4, 10, 11, 6]]

    ############################################################################

    # !!! PLOT TO VERIFY

    def plot_section_data(self, morph, swc_file):
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        print("Plotting verification of sections")

        from snudda.neurons.neuron_morphology import NeuronMorphology
        n = NeuronMorphology(swc_filename=swc_file,
                             verbose=True)
        ax = n.plot_neuron(plot_axon=False)

        for row in morph:
            # Plot secID and secX
            # import pdb
            # pdb.set_trace()
            s = str(row[4]) + "-" + str(round(row[5], 2))
            txt = ax.text(x=row[1], y=row[2], z=row[3], s=s, color="blue")
            print(s + "Press key")
            input('')
            # txt.remove()

        plt.ion()
        plt.show()
        plt.draw()
        plt.pause(0.001)

    ############################################################################

    def write_simulation_config(self, input_list):

        sim_conf = dict([])

        sim_conf["manifest"] = {
            "$BASE_DIR": ".",
            "$NETWORK_DIR": "$BASE_DIR/networks",
            "$COMPONENT_DIR": "$BASE_DIR/components",
            "$INPUT_DIR": "$BASE_DIR/inputs",
            "$OUTPUT_DIR": "$BASE_DIR/output"}

        sim_conf["run"] = {
            "tstop": 1000.0,
            "dt": 0.1,
            "dL": 15,
            "spike_threshold": -15}

        sim_conf["conditions"] = {}

        sim_conf["output"] = {"log_file": "$OUTPUT_DIR/striatum-log.txt",
                              "output_dir": "$OUTPUT_DIR"}

        sim_conf["overwrite_output_dir"] = True

        sim_conf["network"] = "$BASE_DIR/circuit_config.json"

        # node_sets_file contains the reports we want written from the simulation
        # simConf["node_sets_file"] = None # !!! THIS NEEDS TO BE WRITTEN

        if input_list:
            sim_conf["inputs"] = dict()

            for volume_name, sonata_input_hdf5, \
                sonata_virtual_node_hdf5, sonata_virtual_node_csv, \
                    sonata_virtual_edges_hdf5, sonata_virtual_edges_csv in input_list:

                input_info = dict()
                input_info["input_type"] = "spikes"
                input_info["module"] = "h5"
                input_info["input_file"] = f"$INPUT_DIR/{os.path.basename(sonata_input_hdf5)}"
                input_info["node_set"] = f"{volume_name}-input"

                sim_conf["inputs"][f"{volume_name}_spikes"] = input_info

        out_conf_file = os.path.join(self.out_dir, "simulation_config.json")
        print(f"Writing {out_conf_file}")

        with open(out_conf_file, 'wt') as f:
            json.dump(sim_conf, f, indent=4)

    ############################################################################

    def copy_morphologies(self):
        from shutil import copyfile
        import os

        self.morph_location_lookup = dict([])

        print("Copying morphologies")

        morph_set = set([n["morphology"] for n in self.snudda_load.data["neurons"]])

        for morph_path in morph_set:
            morph_file = snudda_parse_path(morph_path, snudda_data=self.snudda_load.data["snudda_data"])
            base_name = os.path.basename(morph_file)
            dest_file = os.path.join(self.out_dir, "components", "morphologies", base_name)

            if self.debug:
                print(f"Copying {morph_file} to {dest_file}")

            copyfile(morph_file, dest_file)
            self.morph_location_lookup[morph_path] = os.path.relpath(dest_file)

    ############################################################################

    def copy_hoc(self):

        self.hoc_location_lookup = dict([])

        print("Copying hoc files...")

        missing_files = []

        hoc_set = set([n["hoc"] for n in self.snudda_load.data["neurons"] if n["hoc"]])

        for hoc_path in hoc_set:
            hoc_file = snudda_parse_path(hoc_path, snudda_data=self.snudda_load.data["snudda_data"])
            if os.path.isfile(hoc_file):
                base_name = os.path.basename(hoc_file)
                dest_file = os.path.join(self.out_dir, "components", "hoc_templates", base_name)

                if self.debug:
                    print(f"Copying {hoc_file} to {dest_file}")

                self.hoc_location_lookup[hoc_path] = dest_file
                copyfile(hoc_file, dest_file)
            else:
                if hoc_file not in missing_files:
                    # Only print same error message ones
                    print(f"- Missing hoc file: {hoc_file}")

                missing_files.append(hoc_file)

    ############################################################################

    def copy_mechanisms(self):

        print("Copying mechanisms")
        mech_path = snudda_parse_path(os.path.join("$SNUDDA_DATA", "neurons", "mechanisms"),
                                      snudda_data=self.snudda_load.data["snudda_data"])

        for mech in glob(os.path.join(mech_path, "*.mod")):
            if self.debug:
                print(f"Copying {mech}")

            mech_file = os.path.basename(mech)
            copyfile(mech, os.path.join(self.out_dir, "components", "mechanisms", mech_file))

    def copy_nest_synapses(self):
        print("Copying NEST synapses")

        mech_path = snudda_parse_path(os.path.join("$SNUDDA_DATA", "nest", "synapses"),
                                      snudda_data=self.snudda_load.data["snudda_data"])

        for mech in glob(os.path.join(mech_path, "*.json")):
            if self.debug:
                print(f"Copying {mech}")

            mech_file = os.path.basename(mech)
            copyfile(mech, os.path.join(self.out_dir, "components", "synapse_dynamics", mech_file))

    ############################################################################

    def sort_input(self, nl, node_type_id_lookup, input_name=None):

        raise DeprecationWarning("Code is deprecated")

        if self.input_file is None or not os.path.isfile(self.input_file):
            print("No input file has been specified!")
            return None

        max_input = 1000000

        input_mat = np.zeros((max_input, 2))
        input_ctr = 0

        if input_name is not None:
            print(f"Processing inputs {input_name}")
        else:
            print("Processing all virtual inputs")

        with h5py.File(self.input_file) as f:

            try:
                for inp in f["input"]:

                    if input_name is not None and input_name != inp:
                        # Does not match required inputName
                        continue

                    if "activity" in f["input"][inp]:
                        # This is a virtual neuron, save it
                        gid = node_type_id_lookup[inp]

                        for spikeTime in f["input"][inp]["activity"]["spikes"]:
                            input_mat[input_ctr, :] = [spikeTime, gid]
                            input_ctr += 1

                            if input_ctr >= max_input:
                                print(f"Expanding input matrix to {max_input}")
                                max_input *= 2
                                new_input_mat = np.zeros((max_input, 2))
                                new_input_mat[:input_ctr, :] = input_mat
                                input_mat = new_input_mat
            except:
                import traceback
                tstr = traceback.format_exc()
                print(tstr)
                import pdb
                pdb.set_trace()

        print("About to sort spikes in order")

        sort_idx = np.argsort(input_mat[:input_ctr, 0])
        input_matrix = input_mat[sort_idx, :]

        return input_matrix

        ############################################################################

    def add_virtual_input(self, volume_name, virtual_node_id, neuron_type, node_id, weight, conv_hurt):

        """ Creates one virtual neurons, to provide external input to that the neuron population

            To know the target neuron, we need to know both neuron population, and node id within that population

            Args:
                volume_name (str): Name of volume
                virtual_node_id (np.array): ID of all virtual nodes
                neuron_type (list of str): Which population does each input target
                node_id (np.array): ID of nodes that are connected
                conv_hurt : ConvHurt object
        """

        virtual_neuron_population = f"{volume_name}-input"
        virtual_node_file = f"{virtual_neuron_population}_nodes.hdf5"

        n_virtual_nodes = len(virtual_node_id)
        node_data = dict()  # Let's see if we get away without specifying coordinates
        virtual_node_type_id = np.zeros(shape=(n_virtual_nodes, ), dtype=int)
        virtual_node_group_id = np.zeros(shape=(n_virtual_nodes, ), dtype=int)
        virtual_node_group_index = np.arange(0, n_virtual_nodes, dtype=int)

        conv_hurt.write_nodes(node_file=virtual_node_file,
                              population_name=virtual_neuron_population,
                              data=node_data,
                              node_id=virtual_node_id,
                              node_type_id=virtual_node_type_id,
                              node_group_id=virtual_node_group_id,
                              node_group_index=virtual_node_group_index,
                              close_file=True)

        csv_virtual_node_types_file = f"{virtual_neuron_population}_node_types.csv"

        virtual_node_data_csv = OrderedDict([("model_type", ["virtual"])])
        conv_hurt.write_node_csv(node_csv_file=csv_virtual_node_types_file,
                                 node_type_id=[0],
                                 data=virtual_node_data_csv)

        edge_file = f"{virtual_neuron_population}_edges.hdf5"

        population_rows = dict()

        # We need to write separate sets of edges for each targeted population
        for nrn_type in set(neuron_type):
            population_rows[f"{volume_name}-input_{nrn_type}"] = \
                np.array([i for (i, val) in enumerate(neuron_type) if val == nrn_type], dtype=int)
            # population_rows[nrn_type] = np.where(np.array(neuron_type) == nrn_type)[0]

        edge_data = {"syn_weight": weight}

        conv_hurt.write_edges(edge_file=edge_file,
                              population_rows=population_rows,
                              edge_type_id=np.zeros(len(virtual_node_id), dtype=int),
                              source_id=virtual_node_id,
                              target_id=node_id,
                              data=edge_data)

        edge_types_file_csv = f"{virtual_neuron_population}_edge_types.csv"
        edge_type_id = np.array([0])
        edge_data_csv = {"model_template": ["static_synapse"]}

        conv_hurt.write_edges_csv(edge_csv_file=edge_types_file_csv,
                                  edge_type_id=edge_type_id,
                                  data=edge_data_csv)

        return virtual_node_file, csv_virtual_node_types_file, edge_file, edge_types_file_csv

    def add_edges_for_virtual_neurons_input(self):

        pass

    ############################################################################


if __name__ == "__main__":

    if len(sys.argv) > 1:
        networkFile = sys.argv[1]
    else:
        networkFile = "last"

    if len(sys.argv) > 2:
        inputFile = sys.argv[2]
    else:
        inputFile = None

    if len(sys.argv) > 3:
        outDir = sys.argv[3]
    else:
        outDir = "striatumSim/"

    cn = ExportSonata(network_file=networkFile,
                      input_file=inputFile,
                      out_dir=outDir)

    print("!!! VERIFY THAT SECTION DATA IS CORRECT")
    # cn.plotSectionData()



import numpy as np
import pandas as pd
from snudda import SnuddaLoad
import matplotlib.pyplot as plt


class PlotConnectivity:

    def __init__(self, network_path=None, network_name="default"):

        self.network_data = dict()
        self.connection_matrix = dict()
        self.distance_matrix = dict()
        self.exp_and_model_data = []
        self.exp_data = None

        if network_path is not None:
            self.add_network(network_path=network_path, network_name=network_name)

    def add_network(self, network_path, network_name):

        """ Load network.

        Args:
            network_path (str) : Path to network
            network_name (str) : Name of network (default: default)"""

        self.network_data[network_name] = SnuddaLoad(network_file=network_path)
        self.connection_matrix[network_name] = self.network_data[network_name].create_connection_matrix()
        self.distance_matrix[network_name] = self.network_data[network_name].create_distance_matrix()

    def add_plot(self,
                 pre_neuron_type, post_neuron_type,
                 network_name="default",
                 plot_label=None,
                 colour="black",
                 error_colour="grey",
                 bin_width=20e-6, num_bins=10,
                 pre_neuron_id=None, post_neuron_id=None,
                 axis=None):

        if pre_neuron_id is None:
            pre_neuron_id = self.network_data[network_name].get_neuron_id_of_type(neuron_type=pre_neuron_type)

        if post_neuron_id is None:
            post_neuron_id = self.network_data[network_name].get_neuron_id_of_type(neuron_type=post_neuron_type)

        sub_con_mat = self.connection_matrix[network_name][pre_neuron_id, :][:, post_neuron_id]
        sub_dist_mat = self.network_data[network_name].create_distance_matrix(pre_id=pre_neuron_id,
                                                                              post_id=post_neuron_id)

        n_connected, n_total, bin_edges = self.matrix_to_bins(con_mat=sub_con_mat, dist_mat=sub_dist_mat,
                                                              bin_width=bin_width, num_bins=num_bins)

        if plot_label is None:
            if network_name == "default":
                plot_label = f"{pre_neuron_type} to {post_neuron_type}"
            else:
                plot_label = f"{network_name}: {pre_neuron_type} to {post_neuron_type}"

        if axis is None:
            fig, axis = plt.subplots()

        p, p_error = self.estimate_p(n_connected=n_connected, n_total=n_total)

        dist = (bin_edges[0:-1] + bin_edges[1:]) / 2 * 1e6

        axis.fill_between(dist, 100*(p-p_error), 100*(p+p_error), color=error_colour, step=None, alpha=0.4)
        axis.plot(dist, 100*p, color=colour, linewidth=2, label=plot_label)

        axis.legend(fontsize=18)

        axis.set_xlabel("Distance between somas ($\mu$m)", fontsize=22)
        axis.set_ylabel("Connection probability (%)", fontsize=22)

        return axis, (n_connected, n_total, dist)

    def estimate_p(self, n_connected, n_total, z=1.96):  # 95 % confidence interval

        p = np.divide(n_connected + (z ** 2) / 2, n_total + z ** 2)
        p_error = np.multiply(np.divide(z, n_total + z ** 2),
                              np.sqrt(np.divide(np.multiply(n_connected, (n_total - n_connected)), n_total)
                                      + (z ** 2)/4))

        return p, p_error

    def draw_plot(self, axis, figure_path=None):

        plt.sca(axis)

        if figure_path is not None:
            plt.savefig(fname=figure_path)

        plt.show()

    def add_experimental_data(self, network_name,
                              pre_neuron_type, post_neuron_type,
                              distance_range, n_connected, n_total,
                              exp_colour="red",
                              model_colour="black",
                              label=None):

        # Calculate the corresponding model data
        pre_neuron_id = self.network_data[network_name].get_neuron_id_of_type(neuron_type=pre_neuron_type)
        post_neuron_id = self.network_data[network_name].get_neuron_id_of_type(neuron_type=post_neuron_type)

        sub_con_mat = self.connection_matrix[network_name][pre_neuron_id, :][:, post_neuron_id]
        sub_dist_mat = self.network_data[network_name].create_distance_matrix(pre_id=pre_neuron_id,
                                                                              post_id=post_neuron_id)

        if len(distance_range) != 2 or distance_range[1] < distance_range[0]:
            raise ValueError("distance_range must be a tuple (start_dist, end_dist) in meters")

        bin_width = distance_range[1] - distance_range[0]
        num_bins = 1

        n_connected_model, n_total_model, bin_edges = self.matrix_to_bins(con_mat=sub_con_mat, dist_mat=sub_dist_mat,
                                                              bin_width=bin_width, num_bins=num_bins)

        if label is None:
            if network_name == "default":
                label = f"{pre_neuron_type} to {post_neuron_type}\n{distance_range[0]*1e6:.0f}-{distance_range[1]*1e6:.0f} $\mu$m"
            else:
                label = f"{pre_neuron_type} to {post_neuron_type} ({network_name})\n{distance_range[0] * 1e6:.0f}-{distance_range[1] * 1e6:.0f} $\mu$m"

        model_label = fr"{label} (model)"
        exp_label = fr"{label}"

        self.exp_and_model_data.append(((n_connected, n_total, exp_label, exp_colour),
                                        (n_connected_model[0], n_total_model[0], model_label, model_colour)))

    def clear_plot_data(self):
        self.exp_and_model_data = []

    def draw_exp_model_data(self, figure_path=None, label_rotation=90):

        fig, axis = plt.subplots()

        n_data_sets = len(self.exp_and_model_data)
        data_set_id = np.arange(n_data_sets)

        label_text = []
        label_pos = []

        for ctr, ((n_con_exp, n_total_exp, exp_label, exp_colour),
                  (n_con_model, n_total_model, model_label, model_colour)) in \
                enumerate(self.exp_and_model_data):

            p_exp, p_exp_error = self.estimate_p(n_connected=n_con_exp, n_total=n_total_exp)
            p_model, p_model_error = self.estimate_p(n_connected=n_con_model, n_total=n_total_model)

            exp_alpha = 1  # 1 - 2*p_exp_error / p_exp
            model_alpha = 1  # 1 - 2*p_model_error / p_model

            axis.bar(ctr-0.15, height=p_exp, yerr=p_exp_error, width=0.3, color=exp_colour, align="center", alpha=exp_alpha, capsize=5)
            axis.bar(ctr+0.15, height=p_model, yerr=p_model_error, width=0.3, color=model_colour, align="center", alpha=model_alpha, capsize=5)

            label_text.append(exp_label)
            label_pos.append(ctr)

        axis.set_xticks(label_pos, label_text, rotation=label_rotation)

        if figure_path is not None:
            plt.savefig(figure_path)

        plt.show()

    @staticmethod
    def matrix_to_bins(con_mat, dist_mat, bin_width, num_bins):

        bin_edges = bin_width * np.arange(0, num_bins+1)
        n_connected = np.zeros((num_bins,), dtype=int)
        n_total = np.zeros((num_bins,), dtype=int)

        for is_connected, distance in zip(con_mat.todense().flatten().T, dist_mat.flatten().T):
            bin_idx = int(distance / bin_width)

            if bin_idx < num_bins:
                n_total[bin_idx] += 1

                if is_connected:
                    n_connected[bin_idx] += 1

        return n_connected, n_total, bin_edges

    def load_exp_data(self, data_file):

        self.exp_data = pd.read_csv(filepath_or_buffer=data_file)

    def add_loaded_exp_data_to_plot(self, phenotype, data_source, network_name,
                                    pre_neuron_type, post_neuron_type,
                                    exp_colour="red", model_colour="black"):

        idx = (self.exp_data["phenotype"] == phenotype) & (self.exp_data["data_source"] == data_source) \
            & (self.exp_data["pre_neuron"] == pre_neuron_type) & (self.exp_data["post_neuron"] == post_neuron_type)

        mean_dist = self.exp_data[idx]["mean_dist"].iloc[0]
        std_dist = self.exp_data[idx]["std_dist"].iloc[0]
        min_dist = self.exp_data[idx]["min_dist"].iloc[0]
        max_dist = self.exp_data[idx]["max_dist"].iloc[0]
        n_connected = self.exp_data[idx]["n_connected"].iloc[0]
        n_total = self.exp_data[idx]["n_total"].iloc[0]

        if not np.isnan(mean_dist) and not np.isnan(std_dist):
            distance_range = np.array([mean_dist - 1.96*std_dist, mean_dist + 1.96*std_dist])
        elif not np.isnan(min_dist) and not np.isnan(max_dist):
            distance_range = np.array([min_dist, max_dist])
        else:
            raise ValueError(f"No usable distances for {phenotype}, {data_source}, {pre_neuron_type}, {post_neuron_type}")

        self.add_experimental_data(network_name=network_name,
                                   pre_neuron_type=pre_neuron_type,
                                   post_neuron_type=post_neuron_type,
                                   distance_range=distance_range,
                                   n_connected=n_connected,
                                   n_total=n_total,
                                   exp_colour=exp_colour,
                                   model_colour=model_colour)
